

import os

from jk_cachefunccalls import cacheCalls
import jk_etcpasswd

from .parsing_utils import *
from .invoke_utils import run





def parse_etc_hostname(stdout:str) -> dict:
	return {
		"hostname": stdout.strip()
	}
#



def get_etc_hostname(c = None) -> dict:
	stdout, stderr, exitcode = run(c, "cat /etc/hostname")

	return parse_etc_hostname(stdout)
#




















import re
import typing
import os
import json

from .invoke_utils import run





def _parsePC(text:str) -> float:
	text = text.strip()

	m = re.match("^(([0-9]+\.[0-9]+)|0)%$", text)
	if not m:
		raise Exception("Failed to parse: " + repr(text))
	return float(m.group(1)) / 100.0
#

def _parseBytes(text:str) -> int:
	text = text.strip()
	orgText = text
	textLC = text.lower()

	factor = 1
	if textLC.endswith("kib"):
		factor = 1024
		text = text[:-3]
	elif textLC.endswith("mib"):
		factor = 1024*1024
		text = text[:-3]
	elif textLC.endswith("gib"):
		factor = 1024*1024*1024
		text = text[:-3]
	elif textLC.endswith("tib"):
		factor = 1024*1024*1024*1024
		text = text[:-3]
	elif textLC.endswith("kb"):
		factor = 1000
		text = text[:-2]
	elif textLC.endswith("mb"):
		factor = 1000000
		text = text[:-2]
	elif textLC.endswith("gb"):
		factor = 1000000000
		text = text[:-2]
	elif textLC.endswith("tb"):
		factor = 1000000000000
		text = text[:-2]
	elif textLC.endswith("b"):
		text = text[:-1]
	
	if not text or not text[-1].isnumeric():
		raise Exception("Parse error: " + repr(orgText))

	value = int(round(float(text) * factor))

	return value
#




def parse_docker_stats(stdout:str, stderr:str, exitcode:int) -> typing.List[typing.Dict[str,typing.Any]]:
	if exitcode != 0:
		raise Exception()

	ret = []

	for line in stdout.split("\n"):
		if not line:
			continue

		jData = json.loads(line)

		containerID = jData["ID"]
		name = jData["Name"]
		nPIDs = int(jData["PIDs"])
		cpuPC1 = _parsePC(jData["CPUPerc"])
		memPC1 = _parsePC(jData["MemPerc"])

		sMemUsage, sMemLimit = jData["MemUsage"].split("/")
		memUsage = _parseBytes(sMemUsage)
		memLimit = _parseBytes(sMemLimit)

		sNetIOr, sNetIOw = jData["NetIO"].split("/")
		netIOr = _parseBytes(sNetIOr)
		netIOw = _parseBytes(sNetIOw)

		sBlockIOr, sBlockIOw = jData["BlockIO"].split("/")
		blockIOr = _parseBytes(sBlockIOr)
		blockIOw = _parseBytes(sBlockIOw)

		ret.append({
			"id": containerID,
			"name": name,
			"pids": nPIDs,
			"memory": {
				"load1": memPC1,
				"usage": memUsage,
				"limit": memLimit,
			},
			"cpu": {
				"load1": cpuPC1,
			},
			"netIO": {
				"read": netIOr,
				"write": netIOw,
			},
			"blockIO": {
				"read": blockIOr,
				"write": blockIOw,
			},
		})

	return ret
#



def get_docker_stats(c = None) -> dict:
	# stdout, stderr, exitcode = run(c, "/usr/bin/docker stats --no-trunc --no-stream")
	stdout, stderr, exitcode = run(c, "/usr/bin/docker stats --no-trunc --no-stream --format '{{json .}}'")
	return parse_docker_stats(stdout, stderr, exitcode)
#



def has_local_docker() -> bool:
	# TODO: can we generalize this?
	return os.path.isfile("/usr/bin/docker")
#












import os
import typing
import sys
import re

import jk_typing
import jk_argparsing
import jk_logging
import jk_console
import jk_flexdata

import jk_sysinfo
import jk_sysinfo.entity






class CollectedSystemData(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self) -> None:
		self.os_release:jk_flexdata.FlexObject = None
		self.bIsRPi:bool = None
		self.data_lsb_release_a:jk_flexdata.FlexObject = None
		self.data_lshw:jk_flexdata.FlexObject = None
		self.data_proccpu_extra:jk_flexdata.FlexObject = None
		self.data_proccpu_extra:jk_flexdata.FlexObject = None
		self.data_mobo:jk_flexdata.FlexObject = None
		self.data_bios:jk_flexdata.FlexObject = None
		self.data_proccpu:jk_flexdata.FlexObject = None
		self.data_cpu:jk_flexdata.FlexObject = None
		self.data_sensors:jk_flexdata.FlexObject = None
		self.data_sysload:jk_flexdata.FlexObject = None
		self.data_mem:jk_flexdata.FlexObject = None
		self.data_lsblk:jk_flexdata.FlexObject = None
		self.data_reboot:jk_flexdata.FlexObject = None
		self.data_mounts:jk_flexdata.FlexObject = None
		self.data_df:jk_flexdata.FlexObject = None
		self.data_net_info:jk_flexdata.FlexObject = None
		self.data_uptime:jk_flexdata.FlexObject = None
		self.data_df:jk_flexdata.FlexObject = None

	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

#






class CLISysInfo(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self):
		pass
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def __createArgsParser(self) -> jk_argparsing.ArgsParser:
		ap = jk_argparsing.ArgsParser(os.path.basename(__file__), "Display system information in a human readable way.")

		ap.optionDataDefaults.set("help", False)
		ap.optionDataDefaults.set("color", True)

		ap.createOption('h', 'help', "Display this help text.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("help", True)
		ap.createOption(None, 'color', "Force using colors in output.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("color", False)
		ap.createOption(None, 'no-color', "Don't use colors in output.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("color", False)

		ap.createAuthor("Jürgen Knauth", "jk@binary-overflow.de")
		ap.setLicense("Apache", YEAR = "2021-2024", COPYRIGHTHOLDER = "Jürgen Knauth")

		ap.createReturnCode(0, "Everything is okay.")
		ap.createReturnCode(1, "An error occurred.")

		return ap
	#

	def __printFlags(self, text:str, flags:typing.Iterable[str]):
		_t = []
		for _c in text:
			if _c == "\t":
				_t.append(_c)
			else:
				_t.append(' ')
		textEmpty = "".join(_t)

		print(text, end="")

		i = 0
		for flag in flags:
			if i == 20:
				# line break
				print(",")
				print(textEmpty, end="")
				i = 0
			else:
				if i > 0:
					# add comma
					print(", ", end="")
				i += 1
			print(flag, end="")
		print()
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	#
	# Collect all data and return it.
	#
	# @param	fabric.Connection c			(optional) A fabric connection object.
	# @return	CollectedSystemData			(always) The collected system data.
	#
	def collectSystemData(self, c = None) -> CollectedSystemData:
		#collectedSystemData = CollectedSystemData()
		ret = CollectedSystemData()

		ret.os_release = jk_flexdata.createFromData(jk_sysinfo.get_etc_os_release(c))

		ret.bIsRPi = ret.os_release.distribution == "raspbian"

		ret.data_lsb_release_a = jk_flexdata.createFromData(jk_sysinfo.get_lsb_release_a(c))				# static
		ret.data_lshw = jk_flexdata.createFromData(jk_sysinfo.get_lshw(c))									# static
		if ret.bIsRPi:
			ret.data_proccpu_extra = jk_flexdata.createFromData(jk_sysinfo.get_proc_cpu_info(c)[1])			# static
			_data_mobo_json = {
				"vendor": "Raspberry Pi Foundation",
			}
			if ret.data_proccpu_extra.hardware:
				_data_mobo_json["soc"] = ret.data_proccpu_extra.hardware
				_data_mobo_json["serial"] = ret.data_proccpu_extra.serial
				_data_mobo_json["name"] = ret.data_proccpu_extra.model
			ret.data_mobo = jk_flexdata.createFromData(_data_mobo_json)
		else:
			ret.data_mobo = jk_flexdata.createFromData(jk_sysinfo.get_motherboard_info(c))					# static
		ret.data_bios = jk_flexdata.createFromData(jk_sysinfo.get_bios_info(c))								# static
		ret.data_proccpu = [ jk_flexdata.createFromData(x) for x in jk_sysinfo.get_proc_cpu_info(c)[0] ]	# static, (runtime)
		ret.data_cpu = jk_flexdata.createFromData(jk_sysinfo.get_cpu_info(c))								# static
		if ret.bIsRPi:
			_t = jk_sysinfo.get_vcgencmd_measure_temp(c)
			_v = jk_sysinfo.get_vcgencmd_measure_volts(c)
			_data_sensors_json = {
				"coretemp": {
					"device": "coretemp",
					"sensorData": {
						"temp1": {
							"sensor": "temp",
							"value": _t["cpu"]["temp"],
						}
					}
				},
				"corevolts": {
					"device": "corevolts",
					"sensorData": {
						"temp1": {
							"sensor": "volt",
							"value": _v["cpu"]["volt"],
						}
					}
				},
				"ramolts": {
					"device": "ramvolts",
					"sensorData": {
						"volt1": {
							"sensor": "volt",
							"value": _v["ram"]["volt"],
						}
					}
				}
			}
			ret.data_sensors = jk_flexdata.createFromData(_data_sensors_json)								# runtime
		else:
			ret.data_sensors = jk_flexdata.createFromData(jk_sysinfo.get_sensors(c))						# runtime
		ret.data_sysload = jk_flexdata.createFromData(jk_sysinfo.get_proc_load_avg(c))						# runtime
		ret.data_mem = jk_flexdata.createFromData(jk_sysinfo.get_proc_meminfo(c))							# runtime
		ret.data_lsblk = jk_flexdata.createFromData(jk_sysinfo.get_lsblk(c))								# runtime
		ret.data_reboot = jk_flexdata.createFromData(jk_sysinfo.get_needs_reboot(c))						# runtime
		ret.data_mounts = jk_flexdata.createFromData(jk_sysinfo.get_mount(c))								# runtime
		ret.data_df = jk_flexdata.createFromData(jk_sysinfo.get_df(c))										# runtime
		ret.data_net_info = jk_flexdata.createFromData(jk_sysinfo.get_net_info(c))							# runtime
		ret.data_uptime = jk_flexdata.createFromData(jk_sysinfo.get_uptime(c))								# runtime

		if not ret.data_reboot:
			print(jk_console.Console.ForeGround.RED + "WARNING: Packet 'needrestart' not installed." + jk_console.Console.RESET)

		return ret
	#

	def printCollectedSystemData(self, csd:CollectedSystemData) -> None:
		assert isinstance(csd, CollectedSystemData)

		################################################################

		print("\n#### bios ####\n")
		print("static")
		print("\tvendor:", csd.data_bios.vendor)
		print("\tversion:", csd.data_bios.version)
		print("\tdate:", csd.data_bios.date)
		print("-")

		################################################################

		print("\n#### motherboard ####\n")
		print("static")
		print("\tvendor:", csd.data_mobo.vendor)
		print("\tname:", csd.data_mobo.name)
		print("\tversion:", csd.data_mobo.version)
		print("-")

		################################################################

		print("\n#### busses and bus devices ####\n")
		print("static")

		def _printPCIStruct(data:jk_flexdata.FlexObject, indent:str=""):
			print(indent + data["class"].upper()
				+ " " + (data.product if data.product else "-")
				+ " (" + data.vendor + ")"
				+ " " + (data.description if data.description else "-")
			)
			if data.children:
				for c in data.children:
					_printPCIStruct(c, indent=indent + "\t")
		#

		_bridge = csd.data_lshw._findR(_class="bridge")
		_printPCIStruct(_bridge, indent="\t")
		print("-")

		################################################################

		print("\n#### system ####\n")
		print("static")
		print("\thostname:", csd.data_lshw.id)	# hostname
		print("\tos distribution:", csd.data_lsb_release_a.distribution)
		print("\tos version:", csd.data_lsb_release_a.version)
		print("\tis LTS version:", csd.data_lsb_release_a.lts)
		print("runtime")
		print("\tprocesses:", csd.data_sysload.processes_total)
		print("\tsystem load:", csd.data_sysload.load1, "/", csd.data_sysload.load5, "/", csd.data_sysload.load15)
		days, hours, minutes, seconds, milliseconds = jk_sysinfo.convertSecondsToHumanReadableDuration(csd.data_uptime.uptimeInSeconds)
		print("\tuptime:", days, "day(s),", hours, "hour(s),", minutes, "minute(s),", seconds, "second(s)")
		if csd.data_reboot.needsReboot:
			updatesRequired = set()
			if csd.data_reboot.updateMicroCodeOrABI:
				updatesRequired.add("CPU or ABI")
			if csd.data_reboot.updateKernel:
				updatesRequired.add("kernel")
			if not updatesRequired:
				raise Exception()
			print(jk_console.Console.ForeGround.ORANGE + "\tUpdate required:", ",".join(updatesRequired) + jk_console.Console.RESET)
		print("-")

		################################################################

		print("\n#### cpu ####\n")
		print("static")
		print("\tvendor:", csd.data_proccpu[0].vendor_id)
		print("\tmodel:", csd.data_proccpu[0].model_name)
		print("\tspeed:", jk_sysinfo.formatFrequencyRangeS(csd.data_cpu.freq_min * 1000000, csd.data_cpu.freq_max * 1000000))
		print("\tcpu family:", csd.data_proccpu[0].cpu_family)
		print("\tcores:", len(csd.data_proccpu), "(hyperthreading)" if ("ht" in csd.data_proccpu[0].flags) else "(no hyperthreading)")
		if "cache_size" in csd.data_proccpu[0]._keys():
			print("\tcpu cache size:", csd.data_proccpu[0].cache_size)
		if csd.data_proccpu[0].flags:
			self.__printFlags("\tflags: ", csd.data_proccpu[0].flags)
		if csd.data_proccpu[0].bugs:
			self.__printFlags("\tbugs: ", csd.data_proccpu[0].bugs)

		bHyperThreading = ("ht" in csd.data_proccpu[0].flags)
		print("\thas hyperthreading:", "yes" if bHyperThreading else "no")

		bHardwareVirtualization = ("vmx" in csd.data_proccpu[0].flags) or ("svm" in csd.data_proccpu[0].flags)
		print("\thas hardware virtualization:", "yes" if bHardwareVirtualization else "no")

		bCryptography = ("aes" in csd.data_proccpu[0].flags) or ("ace" in csd.data_proccpu[0].flags) or ("ace2" in csd.data_proccpu[0].flags)
		print("\thas hardware cryptography:", "yes" if bCryptography else "no")

		print("-")

		################################################################

		print("\n#### memory ####\n")
		print("runtime")
		_mem = csd.data_lshw._findR(id="memory")
		assert _mem.units == "bytes"
		#print("size:", jk_sysinfo.formatBytesS(int(mem.size)))
		print("\tmem total:", jk_sysinfo.formatBytesS(csd.data_mem.MemTotal * 1024))
		print("\tmem available:", jk_sysinfo.formatBytesS(csd.data_mem.MemAvailable * 1024))
		print("\tmem free:", jk_sysinfo.formatBytesS(csd.data_mem.MemFree * 1024))
		print("\tmem buffers:", jk_sysinfo.formatBytesS(csd.data_mem.Buffers * 1024))
		print("\tmem cached:", jk_sysinfo.formatBytesS(csd.data_mem.Cached * 1024))
		print("\tswap total:", jk_sysinfo.formatBytesS(csd.data_mem.SwapTotal * 1024))
		print("\tswap free:", jk_sysinfo.formatBytesS(csd.data_mem.SwapFree * 1024))
		print("\tswap cached:", jk_sysinfo.formatBytesS(csd.data_mem.SwapCached * 1024))
		print("-")

		################################################################

		print("\n#### display ####\n")
		print("static")
		for _display in csd.data_lshw._findAllR(id="display"):
			print("\tvendor:", _display.vendor)
			print("\tproduct:", _display.product)
			print("\tdriver:", _display.configuration.driver)
			print("-")

		################################################################

		print("\n#### storage ####\n")
		print("static")
		#for storage in data_lshw._findAllR(id="storage"):
		#	print("\tvendor:", storage.vendor)
		#	print("\tproduct:", storage.product)
		#	print("\tdescription:", storage.description)
		#	print("\tdriver:", storage.configuration.driver)
		#	print("-")
		#for storage in data_lshw._findAllR(id="cdrom"):
		#	print("\tvendor:", storage.vendor)
		#	print("\tproduct:", storage.product)
		#	print("\tdescription:", storage.description)
		#	print("-")

		_data_lsblk_disks = jk_sysinfo.filter_lsblk_devtree(csd.data_lsblk._toDict(), type="disk")
		_diskTable = jk_console.SimpleTable()
		_diskTable.addRow(
			"device",
			"model",
			"vendor",
			"serial",
			"firmwareRevision",
			"formFactor",
			"diskGranularity",
			"rotationRate",
			"transport",
			"size",
			"uuid",
			"readOnly",
			"rotational",
			"hotplug",
			"NCQ",
			"TRIM",
		).hlineAfterRow = True
		for _jDisk in _data_lsblk_disks:
			_devicePath = _jDisk["dev"]
			_data_hdparam_I = jk_sysinfo.get_hdparm_I(devPath=_devicePath)
			_di = jk_sysinfo.entity.DriveInfo(_jDisk, _data_hdparam_I)
			_diskTable.addRow(
				_di.devicePath,
				_di.model,
				_di.vendor,
				_di.serial,
				_di.firmwareRevision,
				_di.formFactor,
				_di.diskGranularity,
				_di.nominalMediaRotationRate,
				_di.transport,
				_di.size,
				_di.uuid if _di.uuid else "-",
				_di.isReadOnly,
				_di.isRotational,
				_di.isHotplug,
				_di.isNCQSupported,
				_di.isTRIMSupported,
			)	
		_diskTable.print(prefix="\t")
		print("-")

		################################################################

		print("\n#### multimedia ####\n")
		print("static")
		for _multimedia in csd.data_lshw._findAllR(id="multimedia"):
			print("\tvendor:", _multimedia.vendor)
			print("\tproduct:", _multimedia.product)
			print("\tdriver:", _multimedia.configuration.driver)
			print("-")

		################################################################

		print("\n#### network (hardware) ####\n")
		print("static")
		for _network in list(csd.data_lshw._findAllR(id="network")) + list(csd.data_lshw._findAllR(id=re.compile("^network:"))):
			# import jk_json
			# jk_json.prettyPrint(_network._toDict())
			print("\tvendor:", _network.vendor)
			print("\tproduct:", _network.product)
			print("\tdevice:", _network.logicalname)		# network device name
			if _network.configuration.link:
				print("\thas_link:", _network.configuration.link == "yes")
			else:
				print("\thas_link:", "unknown")
			if _network.capabilities.tp or _network.capabilities.ethernet:
				# regular twisted pair network

				if _network.maxSpeedInBitsPerSecond:
					speed, unit = jk_sysinfo.formatBitsPerSecond(_network.maxSpeedInBitsPerSecond)
					print("\tspeed maximum:", speed, unit)					# general speed in bits/s

				if _network.configuration.speed:
					print("\tspeed current:", _network.configuration.speed)			# current speed
				if _network.configuration.duplex:
					print("\tduplex:", _network.configuration.duplex)

			elif _network.configuration.wireless:
				# regular wireless network

				print("\twireless standard:", _network.configuration.wireless)			# "IEEE 802.11"

			else:
				raise Exception("Unknown network type")

			print("\tdescription:", _network.description)
			print("\tdriver:", _network.configuration.driver)
			print("\tmac_addr:", _network.serial)
			print("-")

		################################################################

		print("\n#### sensors ####\n")

		def _formatSensorData(data:jk_flexdata.FlexObject) -> typing.Tuple[str,str,str]:
			if data._isEmpty():
				return "n/a"
			if data.sensor == "volt":
				return str(data.value) + "V"
			if data.sensor == "fan":
				return str(data.value) + " rpm"
			elif data.sensor == "temp":
				_bar, _text = jk_sysinfo.formatTemperatureGraphC2(data.value, data.crit if data.crit else 100)
				if data.crit and data.max:
					_extraText = "max: " + str(data.max) + ", crit: " + str(data.crit) + ""
				else:
					_extraText = "n/a"
				return (_bar, _text, _extraText)
			else:
				raise Exception("Unknown: " + repr(data.sensor))
		#

		print("runtime")
		_table = jk_console.SimpleTable()
		_table.addRow(
			"identifier",
			"visualization",
			"data value",
			"range",
		).hlineAfterRow = True
		for _data in csd.data_sensors._values():
			#jk_json.prettyPrint(data._toDict())
			for sensorItemName, sensorItemStruct in _data.sensorData._items():
				_bar, _text, _extraText = _formatSensorData(sensorItemStruct)
				_table.addRow(
					_data.device + "." + sensorItemName,
					_bar,
					_text,
					_extraText,
				)
		_table.print(prefix="\t")
		print("-")

		################################################################

		print("\n#### network (os) ####\n")
		print("runtime")
		_table = jk_console.SimpleTable()
		_table.addRow(
			"ifname",
			"loop",
			"wlan",
			"mac",
			"mtu",
			"rx pkgs",
			"rx dropped",
			"rx errors",
			"tx pkgs",
			"tx dropped",
			"tx errors",
		).hlineAfterRow = True
		for _networkInterface, _networkInterfaceData in csd.data_net_info._items():
			_table.addRow(
				_networkInterface,
				_networkInterfaceData.is_loop,
				_networkInterfaceData.is_wlan,
				_networkInterfaceData.mac_addr,
				_networkInterfaceData.mtu,
				_networkInterfaceData.rx_packets,
				_networkInterfaceData.rx_dropped,
				_networkInterfaceData.rx_errors,
				_networkInterfaceData.tx_packets,
				_networkInterfaceData.tx_dropped,
				_networkInterfaceData.tx_errors,
			)
		_table.print(prefix="\t")
		print("-")

		################################################################

		print("\n#### drives ####\n")

		print("runtime")

		@jk_typing.checkFunctionSignature()
		def _printDevice(data_lsblk:jk_flexdata.FlexObject, data_mounts:jk_flexdata.FlexObject, data_df:jk_flexdata.FlexObject, indent:str=""):
			if data_lsblk.mountpoint and data_lsblk.mountpoint.startswith("/snap"):
				return
			s = indent + data_lsblk.dev

			if data_lsblk.mountpoint:
				s += " @ "
				s += data_lsblk.mountpoint
				sAdd = " :: "
			else:
				sAdd = " :: "

			if data_lsblk.uuid:
				s += sAdd + repr(data_lsblk.uuid)
				sAdd = " ~ "
			if data_lsblk.fstype:
				s += sAdd + data_lsblk.fstype
				sAdd = " ~ "

			print(s)
			indent += "\t"

			if data_mounts and data_lsblk.mountpoint:
				data_df_2 = data_df._get(data_lsblk.mountpoint)
				#jk_json.prettyPrint(data_mounts._toDict())
				#jk_json.prettyPrint(data_df._toDict())
				if data_df_2:
					print(indent
						+ "total:", jk_sysinfo.formatBytesS(data_df_2.spaceTotal)
						+ ", used:", jk_sysinfo.formatBytesS(data_df_2.spaceUsed)
						+ ", free:", jk_sysinfo.formatBytesS(data_df_2.spaceFree)
						+ ", filled:", jk_sysinfo.formatPercentGraphC(data_df_2.spaceUsed, data_df_2.spaceTotal), jk_sysinfo.formatPercent(data_df_2.spaceUsed, data_df_2.spaceTotal)
						)
					#jk_json.prettyPrint(data_df_2._toDict())
				else:
					print(indent + "Not found: " + data_lsblk.mountpoint)

			if data_lsblk.children:
				for c in data_lsblk.children:
					_printDevice(c, data_mounts, data_df, indent)
		#

		# TODO: drive models

		#print(data_lsblk._keys())
		for _d in csd.data_lsblk.deviceTree:
			_printDevice(_d, csd.data_mounts, csd.data_df, "\t")
			# TODO: list logical drives
		print("-")

		################################################################

		print()

	#

	def main(self) -> int:
		ap = self.__createArgsParser()
		parsedArgs = ap.parse()
		#parsedArgs.dump()

		if parsedArgs.optionData["help"]:
			ap.showHelp()
			return 0

		with jk_logging.wrapMain(bColor=parsedArgs.optionData["color"]) as log:
			# if not parsedArgs.optionData["color"]:
			# 	log = jk_logging.ConsoleLogger.create(logMsgFormatter = jk_logging.DEFAULT_LOG_MESSAGE_FORMATTER, printToStdErr = True)

			# ----

			csd = self.collectSystemData()
			self.printCollectedSystemData(csd)
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

#




"""
from fabric import Connection
import jk_pwdinput

REMOTE_HOST = "<ipaddress>"
REMOTE_PORT = 22
REMOTE_LOGIN = "<login>"
REMOTE_PASSWORD = jk_pwdinput.readpwd("Password for " + REMOTE_LOGIN + "@" + REMOTE_HOST + ": ")
c = Connection(host=REMOTE_HOST, user=REMOTE_LOGIN, port=REMOTE_PORT, connect_kwargs={"password": REMOTE_PASSWORD})
"""
c = None









def main():
	app = CLISysInfo()
	sys.exit(app.main())
#


















from jk_cachefunccalls import cacheCalls

from .parsing_utils import *
from .invoke_utils import run






#
# Note: This function does not work on Raspberry Pi computers.
#
# Returns:
#	{
#		"date": "12/06/2017",
#		"vendor": "American Megatrends Inc.",
#		"version": "1.A0"
#	}
#
@cacheCalls(seconds=3, dependArgs=[0])
def get_bios_info(c = None) -> dict:
	bFail = False

	try:
		date, _, _ = run(c, "cat /sys/devices/virtual/dmi/id/bios_date")
		vendor, _, _ = run(c, "cat /sys/devices/virtual/dmi/id/bios_vendor")
		version, _, _ = run(c, "cat /sys/devices/virtual/dmi/id/bios_version")

		return {
			"date": date.strip(),
			"vendor": vendor.strip(),
			"version": version.strip(),
		}
	except:
		bFail = True

	if bFail:
		return {}
		#raise Exception("Unknown system layout: Can't retrieve required information!")
#












import os

from jk_cachefunccalls import cacheCalls
import jk_etcpasswd

from .parsing_utils import *
from .invoke_utils import run





def parse_etc_passwd(stdoutPasswd:str, stdoutShadow:str) -> dict:
	pwdFile = jk_etcpasswd.PwdFile("/etc/passwd", "/etc/shadow", stdoutPasswd, stdoutShadow)
	return pwdFile.toJSON()
#



def get_etc_passwd(c = None) -> dict:
	if os.geteuid() != 0:
		raise Exception("Must be root!")

	stdoutPasswd, stderrPasswd, exitcodePasswd = run(c, "cat /etc/passwd")
	assert exitcodePasswd == 0
	assert not stderrPasswd

	stdoutShadow, stderrShadow, exitcodeShadow = run(c, "cat /etc/shadow")
	assert exitcodeShadow == 0
	assert not stderrShadow

	return parse_etc_passwd(stdoutPasswd, stdoutShadow)
#




















from jk_cachefunccalls import cacheCalls

from .parsing_utils import *
from .invoke_utils import run



#
# Returns:
#
#	{
#		"sirius": {
#			"user": "sirius",
#			"userID": 1000,
#			"groupID": 1000,
#			"homeDir": "/home/sirius",
#			"shell": "/bin/bash"
#		},
#		...
#	}
#
def parse_user_info_etc_passwd(stdout:str, stderr:str, exitcode:int) -> dict:

	"""
	root:x:0:0:root:/root:/bin/bash
	daemon:x:1:1:daemon:/usr/sbin:/usr/sbin/nologin
	bin:x:2:2:bin:/bin:/usr/sbin/nologin
	sys:x:3:3:sys:/dev:/usr/sbin/nologin
	sync:x:4:65534:sync:/bin:/bin/sync
	games:x:5:60:games:/usr/games:/usr/sbin/nologin
	man:x:6:12:man:/var/cache/man:/usr/sbin/nologin
	lp:x:7:7:lp:/var/spool/lpd:/usr/sbin/nologin
	mail:x:8:8:mail:/var/mail:/usr/sbin/nologin
	news:x:9:9:news:/var/spool/news:/usr/sbin/nologin
	uucp:x:10:10:uucp:/var/spool/uucp:/usr/sbin/nologin
	proxy:x:13:13:proxy:/bin:/usr/sbin/nologin
	www-data:x:33:33:www-data:/var/www:/usr/sbin/nologin
	backup:x:34:34:backup:/var/backups:/usr/sbin/nologin
	list:x:38:38:Mailing List Manager:/var/list:/usr/sbin/nologin
	irc:x:39:39:ircd:/var/run/ircd:/usr/sbin/nologin
	gnats:x:41:41:Gnats Bug-Reporting System (admin):/var/lib/gnats:/usr/sbin/nologin
	nobody:x:65534:65534:nobody:/nonexistent:/usr/sbin/nologin
	systemd-network:x:100:102:systemd Network Management,,,:/run/systemd/netif:/usr/sbin/nologin
	systemd-resolve:x:101:103:systemd Resolver,,,:/run/systemd/resolve:/usr/sbin/nologin
	syslog:x:102:106::/home/syslog:/usr/sbin/nologin
	messagebus:x:103:107::/nonexistent:/usr/sbin/nologin
	_apt:x:104:65534::/nonexistent:/usr/sbin/nologin
	uuidd:x:105:111::/run/uuidd:/usr/sbin/nologin
	avahi-autoipd:x:106:112:Avahi autoip daemon,,,:/var/lib/avahi-autoipd:/usr/sbin/nologin
	usbmux:x:107:46:usbmux daemon,,,:/var/lib/usbmux:/usr/sbin/nologin
	dnsmasq:x:108:65534:dnsmasq,,,:/var/lib/misc:/usr/sbin/nologin
	rtkit:x:109:114:RealtimeKit,,,:/proc:/usr/sbin/nologin
	speech-dispatcher:x:110:29:Speech Dispatcher,,,:/var/run/speech-dispatcher:/bin/false
	whoopsie:x:111:117::/nonexistent:/bin/false
	kernoops:x:112:65534:Kernel Oops Tracking Daemon,,,:/:/usr/sbin/nologin
	saned:x:113:119::/var/lib/saned:/usr/sbin/nologin
	pulse:x:114:120:PulseAudio daemon,,,:/var/run/pulse:/usr/sbin/nologin
	avahi:x:115:122:Avahi mDNS daemon,,,:/var/run/avahi-daemon:/usr/sbin/nologin
	colord:x:116:123:colord colour management daemon,,,:/var/lib/colord:/usr/sbin/nologin
	hplip:x:117:7:HPLIP system user,,,:/var/run/hplip:/bin/false
	geoclue:x:118:124::/var/lib/geoclue:/usr/sbin/nologin
	gnome-initial-setup:x:119:65534::/run/gnome-initial-setup/:/bin/false
	gdm:x:120:125:Gnome Display Manager:/var/lib/gdm3:/bin/false
	sirius:x:1000:1000::/home/sirius:/bin/bash
	cups-pk-helper:x:121:116:user for cups-pk-helper service,,,:/home/cups-pk-helper:/usr/sbin/nologin
	festival:x:122:29::/nonexistent:/usr/sbin/nologin
	timidity:x:123:128:TiMidity++ MIDI sequencer service:/etc/timidity:/usr/sbin/nologin
	sddm:x:124:129:Simple Desktop Display Manager:/var/lib/sddm:/bin/false
	statd:x:125:65534::/var/lib/nfs:/usr/sbin/nologin
	sshd:x:126:65534::/run/sshd:/usr/sbin/nologin
	"""

	if exitcode != 0:
		raise Exception()

	lines = stdout.strip().split("\n")

	ret = {}
	for line in lines:
		userName, x, sUserID, sGroupID, comment, homeDir, shell = line.split(":")
		ret[userName] = {
			"user": userName,
			"userID": int(sUserID),
			"groupID": int(sGroupID),
			"homeDir": None if homeDir == "/nonexistent" else homeDir,
			"shell": None if shell == "/usr/sbin/nologin" else shell,
		}

	return ret
#



#
# Returns:
#
#	{
#		"sirius": {
#			"user": "sirius",
#			"isLocked": False,
#			"hasPassword": True
#		},
#		...
#	}
#
def parse_user_info_etc_shadow(stdout:str, stderr:str, exitcode:int) -> dict:

	"""
	root:$6$iSgU4RbH$....:17970:0:99999:7:::
	daemon:*:17647:0:99999:7:::
	bin:*:17647:0:99999:7:::
	sys:*:17647:0:99999:7:::
	sync:*:17647:0:99999:7:::
	games:*:17647:0:99999:7:::
	man:*:17647:0:99999:7:::
	lp:*:17647:0:99999:7:::
	mail:*:17647:0:99999:7:::
	news:*:17647:0:99999:7:::
	uucp:*:17647:0:99999:7:::
	proxy:*:17647:0:99999:7:::
	www-data:*:17647:0:99999:7:::
	backup:*:17647:0:99999:7:::
	list:*:17647:0:99999:7:::
	irc:*:17647:0:99999:7:::
	gnats:*:17647:0:99999:7:::
	nobody:*:17647:0:99999:7:::
	systemd-network:*:17647:0:99999:7:::
	systemd-resolve:*:17647:0:99999:7:::
	syslog:*:17647:0:99999:7:::
	messagebus:*:17647:0:99999:7:::
	_apt:*:17647:0:99999:7:::
	uuidd:*:17647:0:99999:7:::
	avahi-autoipd:*:17647:0:99999:7:::
	usbmux:*:17647:0:99999:7:::
	dnsmasq:*:17647:0:99999:7:::
	rtkit:*:17647:0:99999:7:::
	speech-dispatcher:!:17647:0:99999:7:::
	whoopsie:*:17647:0:99999:7:::
	kernoops:*:17647:0:99999:7:::
	saned:*:17647:0:99999:7:::
	pulse:*:17647:0:99999:7:::
	avahi:*:17647:0:99999:7:::
	colord:*:17647:0:99999:7:::
	hplip:*:17647:0:99999:7:::
	geoclue:*:17647:0:99999:7:::
	gnome-initial-setup:*:17647:0:99999:7:::
	gdm:*:17647:0:99999:7:::
	sirius:$6$Bf.KreAS$....:18136:0:99999:7:::
	cups-pk-helper:*:17970:0:99999:7:::
	festival:*:17977:0:99999:7:::
	timidity:*:17977:0:99999:7:::
	sddm:*:17977:0:99999:7:::
	statd:*:17977:0:99999:7:::
	sshd:*:17977:0:99999:7:::
	"""

	if exitcode != 0:
		raise Exception()

	lines = stdout.strip().split("\n")

	ret = {}
	for line in lines:
		userName, password, x1, x2, x3, x4, x5, x6, x7 = line.split(":")
		x = {
			"user": userName,
			"isDisabled": password.startswith("!"),
			"hasPassword": (password.startswith("$") or password.startswith("!$")) and (len(password) > 10),
		}
		x["canLogin"] = x["hasPassword"] and not x["isDisabled"]
		ret[userName] = x

	return ret
#



#
# Returns:
#
#	{
#		"sirius": {
#			"group": "sirius",
#			"groupID": 1000,
#			"usersInGroup": []
#		},
#		...
#	}
#
def parse_user_info_etc_group(stdout:str, stderr:str, exitcode:int) -> dict:

	"""
	root:x:0:
	daemon:x:1:
	bin:x:2:
	sys:x:3:
	adm:x:4:syslog,sirius
	tty:x:5:
	disk:x:6:
	lp:x:7:
	mail:x:8:
	news:x:9:
	uucp:x:10:
	man:x:12:
	proxy:x:13:
	kmem:x:15:
	dialout:x:20:
	fax:x:21:
	voice:x:22:
	cdrom:x:24:sirius
	floppy:x:25:
	tape:x:26:
	sudo:x:27:sirius
	audio:x:29:pulse,timidity
	dip:x:30:sirius
	www-data:x:33:
	backup:x:34:
	operator:x:37:
	list:x:38:
	irc:x:39:
	src:x:40:
	gnats:x:41:
	shadow:x:42:
	utmp:x:43:
	video:x:44:
	sasl:x:45:
	plugdev:x:46:sirius
	staff:x:50:
	games:x:60:
	users:x:100:
	nogroup:x:65534:
	systemd-journal:x:101:
	systemd-network:x:102:
	systemd-resolve:x:103:
	input:x:104:
	crontab:x:105:
	syslog:x:106:
	messagebus:x:107:
	netdev:x:108:
	mlocate:x:109:
	ssl-cert:x:110:
	uuidd:x:111:
	avahi-autoipd:x:112:
	bluetooth:x:113:
	rtkit:x:114:
	ssh:x:115:
	lpadmin:x:116:sirius
	whoopsie:x:117:
	scanner:x:118:saned
	saned:x:119:
	pulse:x:120:
	pulse-access:x:121:
	avahi:x:122:
	colord:x:123:
	geoclue:x:124:
	gdm:x:125:
	sirius:x:1000:
	sambashare:x:126:sirius
	_cvsadmin:x:127:
	timidity:x:128:
	sddm:x:129:
	"""

	if exitcode != 0:
		raise Exception()

	lines = stdout.strip().split("\n")

	ret = {}
	for line in lines:
		groupName, x, sGroupID, usersInGroup = line.split(":")
		ret[groupName] = {
			"group": groupName,
			"groupID": int(sGroupID),
			"usersInGroup": [] if not usersInGroup else usersInGroup.split(","),
		}

	return ret
#



#
# Returns:
#
#	{
#		"sirius": {
#				"groupID": 1000,
#				"groups": [
#						"adm",
#						"cdrom",
#						"sudo",
#						"dip",
#						"plugdev",
#						"lpadmin",
#						"sambashare"
#				],
#				"hasPassword": true,
#				"homeDir": "/home/sirius",
#				"isLocked": false,
#				"shell": "/bin/bash",
#				"user": "sirius",
#				"userID": 1000
#		},
#		...
#	}
#
@cacheCalls(seconds=3, dependArgs=[0])
def get_user_info(c = None, catEtcShadowScriptCmd:str = None) -> dict:
	if catEtcShadowScriptCmd is None:
		catEtcShadowScriptCmd = "cat /etc/shadow"

	stdOutPasswd, stdErrPasswd, exitCodePasswd = run(c, "cat /etc/passwd")
	stdOutShadow, stdErrShadow, exitCodeShadow = run(c, "sudo " + catEtcShadowScriptCmd)

	retUsers = joinDictsByKey(
		parse_user_info_etc_passwd(stdOutPasswd, stdErrPasswd, exitCodePasswd),
		parse_user_info_etc_shadow(stdOutShadow, stdErrShadow, exitCodeShadow),
	)

	stdOutGroup, stdErrGroup, exitCodeGroup = run(c, "cat /etc/group")
	groups = parse_user_info_etc_group(stdOutGroup, stdErrGroup, exitCodeGroup)

	for data in retUsers.values():
		data["groups"] = []

	for data in groups.values():
		for user in data["usersInGroup"]:
			retUsers[user]["groups"].append(data["group"])

	return retUsers
#




















import datetime
from typing import NoReturn
import pytz
from dateutil import tz

from jk_cachefunccalls import cacheCalls
import jk_cmdoutputparsinghelper

from .parsing_utils import *
from .invoke_utils import run
from .get_date import *









_VALUE_PARSER_WITH_UNIT = jk_cmdoutputparsinghelper.ValueParser_ByteWithUnit()





class _LineGroup(object):

	def __init__(self, lineList:jk_cmdoutputparsinghelper.LineList) -> None:
		self.title = lineList[0]
		self.lines = lineList.extractFromTo(1)
		if self.lines:
			self.lines.removeCommonPrefix("\t")
	#

	def dump(self):
		print(self.title)
		self.lines.dump(prefix="\t")
	#

	def splitAt(self, lineText:str) -> tuple:
		#for i in range(0, len(self.lines)):
		#	print("-\t", repr(self.lines[i]), repr(lineText), self.lines[i] == lineText)

		n = self.lines.findExact(lineText)
		if n <= 0:
			return None, None

		return (
			self.lines.extractFromTo(0, n),
			self.lines.extractFromTo(n + 1),
		)
	#

#

class _GeneralBlock(object):

	"""
		ATA device, with non-removable media

		'Model Number:       xxxxxxxxxxxxxxxxxxxxxx',
		'Serial Number:      xxxxxxxxxxxx',
		'Firmware Revision:  xxxxxxx',
		'Transport:          Serial, ATA8-AST, SATA 1.0a, SATA II Extensions, SATA Rev 2.5, SATA Rev 2.6, SATA Rev 3.0',
	"""

	def __init__(self, lineGroup:_LineGroup) -> None:
		self.deviceType = lineGroup.title
		_map = lineGroup.lines.convertToMapAtDelimIfPossible(":")
		self.model = _map["Model Number"]
		self.serial = _map["Serial Number"]
		self.firmwareRevision = _map["Firmware Revision"]
		self.transport = _map["Transport"]
	#

	def toJSON(self) -> dict:
		return {
			"model": self.model,
			"serial": self.serial,
			"firmwareRevision": self.firmwareRevision,
			"transport": self.transport,
		}
	#

#

class _DriveConfigurationBlock(object):

	"""
		'Logical\t\tmax\tcurrent',
		'cylinders\t16383\t0',
		'heads\t\t16\t0',
		'sectors/track\t63\t0',
		'--',
		'LBA    user addressable sectors:   268435455',
		'LBA48  user addressable sectors:  1953525168',
		'Logical  Sector size:                   512 bytes',
		'Physical Sector size:                  4096 bytes',
		'Logical Sector-0 offset:                  0 bytes',
		'device size with M = 1024*1024:      953869 MBytes',
		'device size with M = 1000*1000:     1000204 MBytes (1000 GB)',
		'cache/buffer size  = unknown',
		'Form Factor: 2.5 inch',
		'Nominal Media Rotation Rate: Solid State Device',
	"""

	def __init__(self, lineGroup:_LineGroup) -> None:
		#lineGroup.dump()
		_partA, _partB = lineGroup.splitAt("--")

		if _partA is None:
			raise Exception("Parsing error!")

		# ----------------------------------------------------------------
		# parse part A

		assert len(_partA) == 4
		#_cyls = _partA[1].split("\t")
		#_heads = _partA[2].split("\t")
		#_sectors = _partA[3].split("\t")

		# ----------------------------------------------------------------
		# parse part B

		_map = _partB.convertToMapAtDelimIfPossible(":")

		self.maxLBA48Sectors = int(_map.get("LBA48  user addressable sectors")) \
									if "LBA48  user addressable sectors" in _map \
									else None

		self.logSectSize = _VALUE_PARSER_WITH_UNIT(_map.get("Logical  Sector size")) \
									if "Logical  Sector size" in _map \
									else None

		self.phySectSize = _VALUE_PARSER_WITH_UNIT(_map.get("Physical Sector size")) \
									if "Physical Sector size" in _map \
									else None

		self.logSectOfs = _VALUE_PARSER_WITH_UNIT(_map.get("Logical Sector-0 offset")) \
									if "Logical Sector-0 offset" in _map \
									else None

		self.totalSize = _VALUE_PARSER_WITH_UNIT(_map.get("device size with M = 1024*1024")) \
									if "device size with M = 1024*1024" in _map \
									else None

		self.formFactor = _map.get("Form Factor") \
									if "Form Factor" in _map \
									else None

		self.nominalMediaRotationRate = _map.get("Nominal Media Rotation Rate") \
									if "Nominal Media Rotation Rate" in _map \
									else None
	#

	def toJSON(self) -> dict:
		return {
			"maxLBA48Sectors": self.maxLBA48Sectors,
			"logSectSize": self.logSectSize,
			"phySectSize": self.phySectSize,
			"logSectOfs": self.logSectOfs,
			"totalSize": self.totalSize,
			"formFactor": self.formFactor,
			"nominalMediaRotationRate": self.nominalMediaRotationRate,
		}
	#

#

class _CommandsFeaturesBlock(object):

	"""
		'   *\tSMART feature set',
		'    \tSecurity Mode feature set',
		'   *\tPower Management feature set',
		'   *\tWrite cache',
		'   *\tLook-ahead',
		'   *\tWRITE_BUFFER command',
		'   *\tREAD_BUFFER command',
		'   *\tNOP cmd',
		'   *\tDOWNLOAD_MICROCODE',
		'   *\tAdvanced Power Management feature set',
		'   *\t48-bit Address feature set',
		'   *\tMandatory FLUSH_CACHE',
		'   *\tFLUSH_CACHE_EXT',
		'   *\tSMART error logging',
		'   *\tSMART self-test',
		'   *\tGeneral Purpose Logging feature set',
		'   *\tWRITE_{DMA|MULTIPLE}_FUA_EXT',
		'   *\t64-bit World wide name',
		'   *\tWRITE_UNCORRECTABLE_EXT command',
		'   *\t{READ,WRITE}_DMA_EXT_GPL commands',
		'   *\tSegmented DOWNLOAD_MICROCODE',
		'    \tunknown 119[8]',
		'   *\tGen1 signaling speed (1.5Gb/s)',
		'   *\tGen2 signaling speed (3.0Gb/s)',
		'   *\tGen3 signaling speed (6.0Gb/s)',
		'   *\tNative Command Queueing (NCQ)',
		'   *\tPhy event counters',
		'   *\tREAD_LOG_DMA_EXT equivalent to READ_LOG_EXT',
		'   *\tDMA Setup Auto-Activate optimization',
		'    \tDevice-initiated interface power management',
		'   *\tSoftware settings preservation',
		'    \tDevice Sleep (DEVSLP)',
		'   *\tSMART Command Transport (SCT) feature set',
		'   *\tSCT Features Control (AC4)',
		'   *\tSCT Data Tables (AC5)',
		'   *\tSANITIZE_ANTIFREEZE_LOCK_EXT command',
		'   *\tSANITIZE feature set',
		'   *\tCRYPTO_SCRAMBLE_EXT command',
		'   *\tBLOCK_ERASE_EXT command',
		'   *\treserved 69[3]',
		'   *\treserved 69[4]',
		'   *\treserved 69[7]',
		'   *\tDOWNLOAD MICROCODE DMA command',
		'   *\tWRITE BUFFER DMA command',
		'   *\tREAD BUFFER DMA command',
		'   *\tData Set Management TRIM supported (limit 8 blocks)',
	"""

	def __init__(self, lineGroup:_LineGroup) -> None:
		self.__data = {}

		for line in lineGroup.lines:
			parts = line.split("\t")
			assert len(parts) == 2
			partA = parts[0].strip()
			partB = parts[1].strip()
			self.__data[partB] = len(partA) > 0
	#

	def toJSON(self) -> dict:
		return self.__data
	#

#





#
#
#
def parse_hdparm_I(stdout:str, stderr:str, exitcode:int, devPath:str) -> dict:
	lineList = jk_cmdoutputparsinghelper.LineList(stdout.split("\n"))
	#lineList.dump()

	# ----------------------------------------------------------------

	# find the beginning
	nStart = lineList.findExact(devPath + ":")
	if nStart < 0:
		raise Exception("Parse error!")
	nStart += 1
	assert not lineList[nStart]		# this should be an empty line
	nStart += 1

	if len(lineList) == 3:
		return {}

	assert lineList[nStart]			# this should not be an empty line

	# find an empty line after beginning
	nEnd = lineList.findExact("", nStart)
	if nEnd < 0:
		raise Exception("Parse error!")

	# extract that part
	lineList = lineList.extractFromTo(nStart, nEnd)

	# process
	lineList.rightTrimAllLines()

	# convert to _LineGroup[]
	#print("@" * 160)
	lineGroups = []
	for lg in lineList.splitAtRows(lineList.identifyRowsNotStartingWithSpaces()):
		#lg.dump()
		#print("@" * 160)
		lineGroups.append(_LineGroup(lg))
		#lineGroups[-1].dump()

	# do we have enough groups?
	assert len(lineGroups) > 3

	# ----------------------------------------------------------------
	# now begin real parsing

	# the first block is the general block. parse it.
	generalBlock = _GeneralBlock(lineGroups[0])
	#print(generalBlock.toJSON())

	# the second block should be the block "Standards:". skip  it.
	assert lineGroups[1].title == "Standards:"

	# the third block should be the block "Configuration:". parse it.
	assert lineGroups[2].title == "Configuration:"
	driveConfigurationBlock = _DriveConfigurationBlock(lineGroups[2])
	#print(driveConfigurationBlock.toJSON())

	# the forth block should be the block "Capabilities:". skip  it.
	assert lineGroups[3].title == "Capabilities:"

	# the fifth block should be the block "Commands/features:". parse it.
	assert lineGroups[4].title == "Commands/features:"
	featuresBlock = _CommandsFeaturesBlock(lineGroups[4])
	#print(featuresBlock.toJSON())

	# skip the rest for now

	# ----------------------------------------------------------------

	return {
		"general": generalBlock.toJSON(),
		"configuration": driveConfigurationBlock.toJSON(),
		"features": featuresBlock.toJSON(),
	}
#



def where_is(c = None, programName:str = None) -> str:
	assert isinstance(programName, str)

	stdout, stderr, exitcode = run(c, "sudo /usr/bin/whereis " + programName)
	assert stdout.startswith(programName + ":")
	stdout = stdout[len(programName) + 1:].strip()
	listOfLocations = stdout.split(" ")
	for loc in listOfLocations:
		if loc.endswith("/" + programName):
			return loc
	
	raise Exception("Not installed: " + programName)
#



#
# Returns:
#
def get_hdparm_I(c = None, devPath:str = None) -> typing.Union[dict,None]:
	if not devPath:
		raise Exception("Argument 'devPath' is required!")

	programPath = where_is(None, "hdparm")

	stdout, stderr, exitcode = run(c, "sudo {} -I {}".format(programPath, devPath))
	return parse_hdparm_I(stdout, stderr, exitcode, devPath)
#




















import json
import re

from jk_cachefunccalls import cacheCalls

from .parsing_utils import *
from .invoke_utils import run





def _isObj(data:dict, filter:dict) -> bool:
	assert isinstance(data, dict)
	assert isinstance(filter, dict)

	for k, v in filter.items():
		if k in data:
			# 1st attempt
			v2 = data[k]
			if v != v2:
				return False
		else:
			if k.startswith("_"):
				# 2nd attempt
				k = k[1:]
				if k in data:
					v2 = data[k]
					if v != v2:
						return False
				else:
					return False
			else:
				return False
	return True
#

def _findAllR(d:dict, **kwargs):
	assert isinstance(d, dict)

	for key, data in d.items():
		if isinstance(data, (list, tuple)):
			for e in data:
				if isinstance(e, dict):
					if _isObj(e, kwargs):
						yield e
			for e in data:
				if isinstance(e, dict):
					yield from _findAllR(e, **kwargs)
		elif isinstance(data, dict):
			if _isObj(data, kwargs):
				yield data
			yield from _findAllR(data, **kwargs)
#

#
# Returns:
#	{
#		"capabilities": {
#			"vsyscall32": "32-bit processes"
#		},
#		"children": [
#			{
#				"children": [
#					{
#						"claimed": true,
#						"class": "memory",
#						"description": "System memory",
#						"id": "memory",
#						"physid": "0",
#						"size": 32635547648,
#						"units": "bytes"
#					},
#					{
#						"businfo": "cpu@0",
#						"capabilities": {
#							"3dnowprefetch": true,
#							"abm": true,
#							"acpi": "thermal control (ACPI)",
#							"adx": true,
#							"aes": true,
#							...
#						},
#						"capacity": 3900000000,
#						"claimed": true,
#						"class": "processor",
#						"id": "cpu",
#						"physid": "1",
#						"product": "Intel(R) Core(TM) i5-6600 CPU @ 3.30GHz",
#						"size": 799992000,
#						"units": "Hz",
#						"vendor": "Intel Corp.",
#						"width": 64
#					},
#					{
#						"businfo": "pci@0000:00:00.0",
#						"children": [
#							{
#								"businfo": "pci@0000:00:02.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"rom": "extension ROM",
#									"vga_controller": true
#								},
#								"claimed": true,
#								"class": "display",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "i915_bpo",
#									"latency": "0"
#								},
#								"description": "VGA compatible controller",
#								"handle": "PCI:0000:00:02.0",
#								"id": "display",
#								"physid": "2",
#								"product": "HD Graphics 530",
#								"vendor": "Intel Corporation",
#								"version": "06",
#								"width": 64
#							},
#							{
#								"businfo": "pci@0000:00:08.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing"
#								},
#								"class": "generic",
#								"clock": 33000000,
#								"configuration": {
#									"latency": "0"
#								},
#								"description": "System peripheral",
#								"handle": "PCI:0000:00:08.0",
#								"id": "generic:0",
#								"physid": "8",
#								"product": "Xeon E3-1200 v5/v6 / E3-1500 v5 / 6th/7th Gen Core Processor Gaussian Mixture Model",
#								"vendor": "Intel Corporation",
#								"version": "00",
#								"width": 64
#							},
#							{
#								"businfo": "pci@0000:00:14.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"xhci": true
#								},
#								"claimed": true,
#								"class": "bus",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "xhci_hcd",
#									"latency": "0"
#								},
#								"description": "USB controller",
#								"handle": "PCI:0000:00:14.0",
#								"id": "usb",
#								"physid": "14",
#								"product": "100 Series/C230 Series Chipset Family USB 3.0 xHCI Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 64
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:17.0",
#								"capabilities": {
#									"ahci_1.0": true,
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"storage": true
#								},
#								"claimed": true,
#								"class": "storage",
#								"clock": 66000000,
#								"configuration": {
#									"driver": "ahci",
#									"latency": "0"
#								},
#								"description": "SATA controller",
#								"handle": "PCI:0000:00:17.0",
#								"id": "storage",
#								"physid": "17",
#								"product": "Q170/Q150/B150/H170/H110/Z170/CM236 Chipset SATA Controller [AHCI Mode]",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:1f.2",
#								"capabilities": {
#									"bus_master": "bus mastering"
#								},
#								"class": "memory",
#								"clock": 33000000,
#								"configuration": {
#									"latency": "0"
#								},
#								"description": "Memory controller",
#								"handle": "PCI:0000:00:1f.2",
#								"id": "memory",
#								"physid": "1f.2",
#								"product": "100 Series/C230 Series Chipset Family Power Management Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							},
#							{
#								"businfo": "pci@0000:00:1f.3",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing"
#								},
#								"claimed": true,
#								"class": "multimedia",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "snd_hda_intel",
#									"latency": "32"
#								},
#								"description": "Audio device",
#								"handle": "PCI:0000:00:1f.3",
#								"id": "multimedia",
#								"physid": "1f.3",
#								"product": "100 Series/C230 Series Chipset Family HD Audio Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 64
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:1f.6",
#								"capabilities": {
#									"1000bt-fd": "1Gbit/s (full duplex)",
#									"100bt": "100Mbit/s",
#									"100bt-fd": "100Mbit/s (full duplex)",
#									"10bt": "10Mbit/s",
#									"10bt-fd": "10Mbit/s (full duplex)",
#									"autonegotiation": "Auto-negotiation",
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"ethernet": true,
#									"physical": "Physical interface",
#									"tp": "twisted pair"
#								},
#								"capacity": 1000000000,
#								"claimed": true,
#								"class": "network",
#								"clock": 33000000,
#								"configuration": {
#									"autonegotiation": "on",
#									"broadcast": "yes",
#									"driver": "e1000e",
#									"driverversion": "3.2.6-k",
#									"duplex": "full",
#									"firmware": "0.7-4",
#									"latency": "0",
#									"link": "yes",
#									"multicast": "yes",
#									"port": "twisted pair",
#									"speed": "1Gbit/s"
#								},
#								"description": "Ethernet interface",
#								"handle": "PCI:0000:00:1f.6",
#								"id": "network",
#								"logicalname": "enp0s31f6",
#								"physid": "1f.6",
#								"product": "Ethernet Connection (2) I219-V",
#								"serial": "d8:cb:8a:ec:5f:05",
#								"size": 1000000000,
#								"units": "bit/s",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							}
#						],
#						"claimed": true,
#						"class": "bridge",
#						"clock": 33000000,
#						"description": "Host bridge",
#						"handle": "PCIBUS:0000:00",
#						"id": "pci",
#						"physid": "100",
#						"product": "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers",
#						"vendor": "Intel Corporation",
#						"version": "07",
#						"width": 32
#					},
#					{
#						"capabilities": {
#							"emulated": "Emulated device"
#						},
#						"children": [
#							{
#								"businfo": "scsi@3:0.0.0",
#								"capabilities": {
#									"audio": "Audio CD playback",
#									"cd-r": "CD-R burning",
#									"cd-rw": "CD-RW burning",
#									"dvd": "DVD playback",
#									"dvd-r": "DVD-R burning",
#									"dvd-ram": "DVD-RAM burning",
#									"removable": "support is removable"
#								},
#								"children": [
#									{
#										"claimed": true,
#										"class": "disk",
#										"configuration": {
#											"mount.fstype": "iso9660",
#											"mount.options": "ro,nosuid,nodev,relatime,uid=1000,gid=1000,iocharset=utf8,mode=0400,dmode=0500",
#											"state": "mounted"
#										},
#										"dev": "11:0",
#										"id": "medium",
#										"logicalname": [
#											"/dev/cdrom",
#											"/media/xxxxxxxx/YYYYYY"
#										],
#										"physid": "0"
#									}
#								],
#								"claimed": true,
#								"class": "disk",
#								"configuration": {
#									"ansiversion": "5",
#									"mount.fstype": "iso9660",
#									"mount.options": "ro,nosuid,nodev,relatime,uid=1000,gid=1000,iocharset=utf8,mode=0400,dmode=0500",
#									"state": "mounted",
#									"status": "ready"
#								},
#								"description": "DVD-RAM writer",
#								"dev": "11:0",
#								"handle": "SCSI:03:00:00:00",
#								"id": "cdrom",
#								"logicalname": [
#									"/dev/cdrom",
#									"/dev/cdrw",
#									"/dev/dvd",
#									"/dev/dvdrw",
#									"/dev/sr0",
#								],
#								"physid": "0.0.0",
#								"product": "CDDVDW SH-S203P",
#								"vendor": "TSSTcorp",
#								"version": "SB00"
#							}
#						],
#						"claimed": true,
#						"class": "storage",
#						"id": "scsi",
#						"logicalname": "scsi3",
#						"physid": "2"
#					}
#				],
#				"claimed": true,
#				"class": "bus",
#				"description": "Motherboard",
#				"id": "core",
#				"physid": "0"
#			}
#		],
#		"claimed": true,
#		"class": "system",
#		"description": "Computer",
#		"id": "nbxxxxxxxx",
#		"width": 64
#	}
#
def parse_lshw(stdout:str, stderr:str, exitcode:int) -> dict:
	try:
		data_lshw = json.loads(stdout)
	except json.decoder.JSONDecodeError as ee:
		raise Exception("JSON parsing error. Please upgrade lshw as your OS seems to use a very old version of lshw.")

	if isinstance(data_lshw, list):
		assert len(data_lshw) == 1
		data_lshw = data_lshw[0]
	assert isinstance(data_lshw, dict)

	# enrich with additional information: network

	for network in _findAllR(data_lshw, id="network"):
		if ("capabilities" in network) and network["capabilities"].get("tp"):
			# regular twisted pair network

			maxSpeedInBitsPerSecond = None
			for key in network["capabilities"].keys():
				m = re.match(r"^(\d+)bt(-fd)?$", key)
				if m:
					x = int(m.groups()[0]) * 1000000
					if (maxSpeedInBitsPerSecond is None) or (x > maxSpeedInBitsPerSecond):
						maxSpeedInBitsPerSecond = x
			if maxSpeedInBitsPerSecond is None:
				if network.get("size"):
					maxSpeedInBitsPerSecond = int(network["size"])

			if maxSpeedInBitsPerSecond:
				network["maxSpeedInBitsPerSecond"] = maxSpeedInBitsPerSecond

	# enrich with additional information: cpu

	for cpu in _findAllR(data_lshw, id="cpu"):
		if "capabilities" in cpu:
			cpu["hyperthreading"] = "ht" in cpu["capabilities"]
			cpu["virtualization"] = "vmx" in cpu["capabilities"]
			cpu["bitArch"] = 64 if "x86-64" in cpu["capabilities"] else 32
			cpu["encryption"] = "aes" in cpu["capabilities"]

	# return data

	return data_lshw
#



#
# Returns:
#	{
#		"capabilities": {
#			"vsyscall32": "32-bit processes"
#		},
#		"children": [
#			{
#				"children": [
#					{
#						"claimed": true,
#						"class": "memory",
#						"description": "System memory",
#						"id": "memory",
#						"physid": "0",
#						"size": 32635547648,
#						"units": "bytes"
#					},
#					{
#						"businfo": "cpu@0",
#						"capabilities": {
#							"3dnowprefetch": true,
#							"abm": true,
#							"acpi": "thermal control (ACPI)",
#							"adx": true,
#							"aes": true,
#							...
#						},
#						"capacity": 3900000000,
#						"claimed": true,
#						"class": "processor",
#						"id": "cpu",
#						"physid": "1",
#						"product": "Intel(R) Core(TM) i5-6600 CPU @ 3.30GHz",
#						"size": 799992000,
#						"units": "Hz",
#						"vendor": "Intel Corp.",
#						"width": 64
#					},
#					{
#						"businfo": "pci@0000:00:00.0",
#						"children": [
#							{
#								"businfo": "pci@0000:00:02.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"rom": "extension ROM",
#									"vga_controller": true
#								},
#								"claimed": true,
#								"class": "display",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "i915_bpo",
#									"latency": "0"
#								},
#								"description": "VGA compatible controller",
#								"handle": "PCI:0000:00:02.0",
#								"id": "display",
#								"physid": "2",
#								"product": "HD Graphics 530",
#								"vendor": "Intel Corporation",
#								"version": "06",
#								"width": 64
#							},
#							{
#								"businfo": "pci@0000:00:08.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing"
#								},
#								"class": "generic",
#								"clock": 33000000,
#								"configuration": {
#									"latency": "0"
#								},
#								"description": "System peripheral",
#								"handle": "PCI:0000:00:08.0",
#								"id": "generic:0",
#								"physid": "8",
#								"product": "Xeon E3-1200 v5/v6 / E3-1500 v5 / 6th/7th Gen Core Processor Gaussian Mixture Model",
#								"vendor": "Intel Corporation",
#								"version": "00",
#								"width": 64
#							},
#							{
#								"businfo": "pci@0000:00:14.0",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"xhci": true
#								},
#								"claimed": true,
#								"class": "bus",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "xhci_hcd",
#									"latency": "0"
#								},
#								"description": "USB controller",
#								"handle": "PCI:0000:00:14.0",
#								"id": "usb",
#								"physid": "14",
#								"product": "100 Series/C230 Series Chipset Family USB 3.0 xHCI Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 64
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:17.0",
#								"capabilities": {
#									"ahci_1.0": true,
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"storage": true
#								},
#								"claimed": true,
#								"class": "storage",
#								"clock": 66000000,
#								"configuration": {
#									"driver": "ahci",
#									"latency": "0"
#								},
#								"description": "SATA controller",
#								"handle": "PCI:0000:00:17.0",
#								"id": "storage",
#								"physid": "17",
#								"product": "Q170/Q150/B150/H170/H110/Z170/CM236 Chipset SATA Controller [AHCI Mode]",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:1f.2",
#								"capabilities": {
#									"bus_master": "bus mastering"
#								},
#								"class": "memory",
#								"clock": 33000000,
#								"configuration": {
#									"latency": "0"
#								},
#								"description": "Memory controller",
#								"handle": "PCI:0000:00:1f.2",
#								"id": "memory",
#								"physid": "1f.2",
#								"product": "100 Series/C230 Series Chipset Family Power Management Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							},
#							{
#								"businfo": "pci@0000:00:1f.3",
#								"capabilities": {
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing"
#								},
#								"claimed": true,
#								"class": "multimedia",
#								"clock": 33000000,
#								"configuration": {
#									"driver": "snd_hda_intel",
#									"latency": "32"
#								},
#								"description": "Audio device",
#								"handle": "PCI:0000:00:1f.3",
#								"id": "multimedia",
#								"physid": "1f.3",
#								"product": "100 Series/C230 Series Chipset Family HD Audio Controller",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 64
#							},
#							...,
#							{
#								"businfo": "pci@0000:00:1f.6",
#								"capabilities": {
#									"1000bt-fd": "1Gbit/s (full duplex)",
#									"100bt": "100Mbit/s",
#									"100bt-fd": "100Mbit/s (full duplex)",
#									"10bt": "10Mbit/s",
#									"10bt-fd": "10Mbit/s (full duplex)",
#									"autonegotiation": "Auto-negotiation",
#									"bus_master": "bus mastering",
#									"cap_list": "PCI capabilities listing",
#									"ethernet": true,
#									"physical": "Physical interface",
#									"tp": "twisted pair"
#								},
#								"capacity": 1000000000,
#								"claimed": true,
#								"class": "network",
#								"clock": 33000000,
#								"configuration": {
#									"autonegotiation": "on",
#									"broadcast": "yes",
#									"driver": "e1000e",
#									"driverversion": "3.2.6-k",
#									"duplex": "full",
#									"firmware": "0.7-4",
#									"latency": "0",
#									"link": "yes",
#									"multicast": "yes",
#									"port": "twisted pair",
#									"speed": "1Gbit/s"
#								},
#								"description": "Ethernet interface",
#								"handle": "PCI:0000:00:1f.6",
#								"id": "network",
#								"logicalname": "enp0s31f6",
#								"physid": "1f.6",
#								"product": "Ethernet Connection (2) I219-V",
#								"serial": "d8:cb:8a:ec:5f:05",
#								"size": 1000000000,
#								"units": "bit/s",
#								"vendor": "Intel Corporation",
#								"version": "31",
#								"width": 32
#							}
#						],
#						"claimed": true,
#						"class": "bridge",
#						"clock": 33000000,
#						"description": "Host bridge",
#						"handle": "PCIBUS:0000:00",
#						"id": "pci",
#						"physid": "100",
#						"product": "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers",
#						"vendor": "Intel Corporation",
#						"version": "07",
#						"width": 32
#					},
#					{
#						"capabilities": {
#							"emulated": "Emulated device"
#						},
#						"children": [
#							{
#								"businfo": "scsi@3:0.0.0",
#								"capabilities": {
#									"audio": "Audio CD playback",
#									"cd-r": "CD-R burning",
#									"cd-rw": "CD-RW burning",
#									"dvd": "DVD playback",
#									"dvd-r": "DVD-R burning",
#									"dvd-ram": "DVD-RAM burning",
#									"removable": "support is removable"
#								},
#								"children": [
#									{
#										"claimed": true,
#										"class": "disk",
#										"configuration": {
#											"mount.fstype": "iso9660",
#											"mount.options": "ro,nosuid,nodev,relatime,uid=1000,gid=1000,iocharset=utf8,mode=0400,dmode=0500",
#											"state": "mounted"
#										},
#										"dev": "11:0",
#										"id": "medium",
#										"logicalname": [
#											"/dev/cdrom",
#											"/media/xxxxxxxx/YYYYYY"
#										],
#										"physid": "0"
#									}
#								],
#								"claimed": true,
#								"class": "disk",
#								"configuration": {
#									"ansiversion": "5",
#									"mount.fstype": "iso9660",
#									"mount.options": "ro,nosuid,nodev,relatime,uid=1000,gid=1000,iocharset=utf8,mode=0400,dmode=0500",
#									"state": "mounted",
#									"status": "ready"
#								},
#								"description": "DVD-RAM writer",
#								"dev": "11:0",
#								"handle": "SCSI:03:00:00:00",
#								"id": "cdrom",
#								"logicalname": [
#									"/dev/cdrom",
#									"/dev/cdrw",
#									"/dev/dvd",
#									"/dev/dvdrw",
#									"/dev/sr0",
#									"/media/xxxxxxxx/YYYYYY"
#								],
#								"physid": "0.0.0",
#								"product": "CDDVDW SH-S203P",
#								"vendor": "TSSTcorp",
#								"version": "SB00"
#							}
#						],
#						"claimed": true,
#						"class": "storage",
#						"id": "scsi",
#						"logicalname": "scsi3",
#						"physid": "2"
#					}
#				],
#				"claimed": true,
#				"class": "bus",
#				"description": "Motherboard",
#				"id": "core",
#				"physid": "0"
#			}
#		],
#		"claimed": true,
#		"class": "system",
#		"description": "Computer",
#		"id": "nbxxxxxxxx",
#		"width": 64
#	}
#
@cacheCalls(seconds=3, dependArgs=[0])
def get_lshw(c = None) -> dict:
	try:
		stdout, stderr, exitcode = run(c, "/usr/bin/lshw -json")
	except:
		# lshw not installed!
		return {}
	return parse_lshw(stdout, stderr, exitcode)
#













__author__ = "Jürgen Knauth"
__version__ = "0.2024.9.18.3"



from	.parsing_utils				import	joinDictsByKey
from	.invoke_utils				import	enableDebugging, run
from	.value_formatting			import	*

from	.get_accesspoints			import																						get_accesspoints
from	.get_apt_list_upgradable	import	parse_apt_list_upgradable,															get_apt_list_upgradable
from	.get_bios_info				import																						get_bios_info
from	.get_cpu_info				import																						get_cpu_info
from	.get_dpkg_list				import	parse_dpkg_list,																	get_dpkg_list
from	.get_date					import	parse_date_as_datetime,																get_date, get_date_as_datetime
from	.get_df						import	parse_df,																			get_df
from	.get_etc_group				import	parse_etc_group,																	get_etc_group
from	.get_etc_hostname			import	parse_etc_hostname,																	get_etc_hostname
from	.get_etc_os_release			import	parse_etc_os_release,																get_etc_os_release
from	.get_etc_passwd				import	parse_etc_passwd,																	get_etc_passwd
from	.get_ifconfig				import	parse_ifconfig,																		get_ifconfig
from	.get_lsb_release_a			import	parse_lsb_release_a,																get_lsb_release_a
from	.get_lsblk					import	parse_lsblk, filter_lsblk_devtree,													get_lsblk
from	.get_lshw					import	parse_lshw,																			get_lshw
from	.get_motherboard_info		import																						get_motherboard_info
from	.get_mount					import	parse_mount,																		get_mount
from	.get_needs_reboot			import	parse_needs_reboot,																	get_needs_reboot
from	.get_net_info				import																						get_net_info
from	.get_pip3_list				import	parse_pip3_list,																	get_pip3_list
from	.get_proc_cpu_info			import	parse_proc_cpu_info,																get_proc_cpu_info
from	.get_proc_load_avg			import	parse_proc_load_avg,																get_proc_load_avg
from	.get_proc_meminfo			import	parse_proc_meminfo,																	get_proc_meminfo
from	.get_ps						import	parse_ps,																			get_ps
from	.get_ps_local				import																						get_ps_local, get_process_info_local
from	.get_sensors				import	parse_sensors,																		get_sensors, has_local_sensors
from	.get_uptime					import	parse_uptime,																		get_uptime
from	.get_user_info				import	parse_user_info_etc_passwd, parse_user_info_etc_shadow, parse_user_info_etc_group,	get_user_info
from	.get_vcgencmd				import																						get_vcgencmd_get_config, get_vcgencmd_measure_volts, get_vcgencmd_measure_temp, get_vcgencmd_get_mem, get_vcgencmd_display_power, get_vcgencmd, has_local_vcgencmd

from	.get_hdparm_I				import	parse_hdparm_I,																		get_hdparm_I
from	.get_dmi_decode				import	parse_dmi_decode,																	get_dmi_decode, DMIDecodeConstants

from	.get_systemctl_units		import	parse_systemctl_units,																get_systemctl_units

from	.get_docker_stats			import parse_docker_stats,																	get_docker_stats, has_local_docker






import re
import os

import jk_cmdoutputparsinghelper
from jk_cachefunccalls import cacheCalls
#import jk_json

from .parsing_utils import *
from .invoke_utils import run





class DMIDecodeConstants:

	BIOS = 0
	SYSTEM = 1
	BASE_BOARD = 2
	CHASSIS = 3
	PROCESSOR = 4
	MEMORY_CONTROLLER = 5
	MEMORY_MODULE = 6
	CPU_CACHE = 7
	PORT_CONNECTOR = 8
	SYSTEM_SLOTS = 9
	ON_BOARD_DEVICES = 10
	OEM_STRINGS = 11
	SYSTEM_CONFIGURATION_OPTIONS = 12
	BIOS_LANGUAGE = 13
	GROUP_ASSOCIATIONS = 14
	SYSTEM_EVENT_LOG = 15
	PHYSICAL_MEMORY_ARRAY = 16
	MEMORY_DEVICE = 17
	MEMORY_ERROR_32_BIT = 18
	MEMORY_ARRAY_MAPPED_ADDRESS = 19
	MEMORY_DEVICE_MAPPED_ADDRESS = 20
	BUILT_IN_POINTING_DEVICE = 21
	PORTABLE_BATTERY = 22
	SYSTEM_RESET = 23
	HARDWARE_SECURITY = 24
	SYSTEM_POWER_CONTROLS = 25
	VOLTAGE_PROBE = 26
	COOLING_DEVICE = 27
	TEMPERATURE_PROBE = 28
	ELECTRICAL_CURRENT_PROBE = 29
	OUT_OF_BAND_REMOTE_ACCESS = 30
	BOOT_INTEGRITY_SERVICES = 31
	SYSTEM_BOOT = 32
	MEMORY_ERROR_64_BIT = 33
	MANAGEMENT_DEVICE = 34
	MANAGEMENT_DEVICE_COMPONENT = 35
	MANAGEMENT_DEVICE_THRESHOLD_DATA = 36
	MEMORY_CHANNEL = 37
	IPMI_DEVICE = 38
	POWER_SUPPLY = 39
	ADDITIONAL_INFORMATION = 40
	ONBOARD_DEVICE = 41

#






def _parse_block(lineList:jk_cmdoutputparsinghelper.LineList) -> dict:
	assert isinstance(lineList, jk_cmdoutputparsinghelper.LineList)

	# lineList.dump()

	assert not lineList[0].startswith("\t")
	assert not lineList[1].startswith("\t")
	if len(lineList) == 2:
		# empty
		return {}
	assert lineList[2].startswith("\t")

	m = re.match("^Handle 0x([0-9A-Fa-f]+), DMI type ([0-9]+), .*", lineList[0])
	if m is None:
		raise Exception("Parse error: " + repr(lineList[0]))

	retData = {}
	ret = {
		"handle": m.group(1),
		"blockType": lineList[1],
		"data": retData,
	}

	# extract data

	lineList = lineList.extractFromTo(2)
	if not lineList.hasCommonPrefix("\t"):
		raise Exception("Parse error: \\t is not a common prefix")
	lineList.removeCommonPrefix("\t")

	# parse data

	_lastKey = None
	_lastValue = None
	_currentList = None
	for line in lineList:
		if line.startswith("\t"):
			# indented line
			if _currentList is not None:
				_currentList.append(line.strip())
			else:
				if _lastKey == "Items":
					# "Items" seems to be a marker for a list, though there is a value
					_currentList = [ line.strip() ]
					retData["Items"] = _currentList
				else:
					# ignore lines that belong to a list but where the kast key had a value
					#print("WARN: " + repr(line))
					#print("\t_lastKey: " + repr(_lastKey))
					#print("\t_lastValue: " + repr(_lastValue))
					pass
			_lastValue = None
		else:
			# regular data line
			pos = line.find(":")
			if pos < 0:
				raise Exception("Parse error: " + repr(line))
			_key = line[:pos]
			_value = line[pos+1:].strip()
			if _value:
				# we have a real value -> we typically don't have a list
				#print("#>>value#", repr(_key), repr(_value), len(_value))
				retData[_key] = _value
				_currentList = None
			else:
				# value is empty -> we have a list
				#print("#>>list#", repr(_key))
				_currentList = []
				retData[_key] = _currentList
				_value = None
			_lastKey = _key
			_lastValue = _value

	# ----

	#jk_json.prettyPrint(ret)
	return ret
#


#
#
#
def parse_dmi_decode(stdout:str, stderr:str, exitcode:int) -> typing.Union[list,None]:
	lineList = jk_cmdoutputparsinghelper.LineList(stdout.split("\n"))
	#lineList.dump()

	lineList.rightTrimAllLines()

	assert lineList
	assert lineList[0].startswith("# dmidecode ")
	if lineList[1].find("sorry") >= 0:
		# no dmidecode available
		return None

	# find empty section
	pos = lineList.findExact("")
	if pos < 0:
		raise Exception("Parse error!")
	lineList = lineList.extractFromTo(pos + 1)
	lineList.removeLeadingEmptyLines()
	lineList.removeTrailingEmptyLines()

	if len(lineList) == 0:
		# empty
		return []

	# ----------------------------------------------------------------

	ret = []
	blocks = lineList.splitAtEmptyLines()
	for block in blocks:
		ret.append(_parse_block(block))
 
	return ret
# 






def get_dmi_decode(c = None, typeNumberOrNumbers:typing.Union[int,typing.List[int],typing.Tuple[int]] = None) -> typing.Union[list,None]:
	if isinstance(typeNumberOrNumbers, (list,tuple)):
		sTypeNos = ",".join([ str(x) for x in typeNumberOrNumbers ])
	else:
		assert isinstance(typeNumberOrNumbers, int)
		sTypeNos = str(typeNumberOrNumbers)

	stdout, stderr, exitcode = run(c, "sudo /usr/sbin/dmidecode --type " + sTypeNos)
	return parse_dmi_decode(stdout, stderr, exitcode)
#















from jk_cachefunccalls import cacheCalls

from jk_cmdoutputparsinghelper import LineList, Table, ColumnDef

from .parsing_utils import *
from .invoke_utils import run




def _parseMark(s:str) -> bool:
	return bool(s)
#




#
# Returns something like this:
#
#	{
#		"automounts": {
#			"proc-sys-fs-binfmt_misc": [
#				false,
#				"proc-sys-fs-binfmt_misc.automount",
#				"loaded",
#				"active",
#				"running",
#				"Arbitrary Executable File Formats File System Automount Point"
#			]
#			....
#		},
#		"devices": {
#			"dev-cdrom": [
#				false,
#				"dev-cdrom.device",
#				"loaded",
#				"active",
#				"plugged",
#				"TSSTcorp_CDDVDW_SH-S203P Gal_Civ_2_DE"
#			],
#			....
#		},
#		"mounts": {
#			"-": [
#				false,
#				"-.mount",
#				"loaded",
#				"active",
#				"mounted",
#				"/"
#			],
#			....
#			"mounts-ramdisk": [
#				false,
#				"mounts-ramdisk.mount",
#				"loaded",
#				"active",
#				"mounted",
#				"/mounts/ramdisk"
#			],
#			....
#		},
#		"paths": {
#			"acpid": [
#				false,
#				"acpid.path",
#				"loaded",
#				"active",
#				"running",
#				"ACPI Events Check"
#			],
#			....
#		},
#		"scopes": {
#			"init": [
#				false,
#				"init.scope",
#				"loaded",
#				"active",
#				"running",
#				"System and Service Manager"
#			],
#			"session-c2": [
#				false,
#				"session-c2.scope",
#				"loaded",
#				"active",
#				"running",
#				"Session c2 of user woodoo"
#			]
#			....
#		},
#		"services": {
#			"accounts-daemon": [
#				false,
#				"accounts-daemon.service",
#				"loaded",
#				"active",
#				"running",
#				"Accounts Service"
#			],
#			....
#			"colord": [
#				false,
#				"colord.service",
#				"loaded",
#				"active",
#				"running",
#				"Manage, Install and Generate Color Profiles"
#			],
#			....
#			"whoopsie": [
#				false,
#				"whoopsie.service",
#				"loaded",
#				"active",
#				"running",
#				"crash report submission daemon"
#			]
#		},
#		"slices": {
#			....
#		},
#		"sockets": {
#			"acpid": [
#				false,
#				"acpid.socket",
#				"loaded",
#				"active",
#				"running",
#				"ACPID Listen Socket"
#			],
#			....
#			"cups": [
#				false,
#				"cups.socket",
#				"loaded",
#				"active",
#				"running",
#				"CUPS Scheduler"
#			],
#			....
#		},
#		"targets": {
#			"basic": [
#				false,
#				"basic.target",
#				"loaded",
#				"active",
#				"active",
#				"Basic System"
#			],
#			....
#			"graphical": [
#				false,
#				"graphical.target",
#				"loaded",
#				"active",
#				"active",
#				"Graphical Interface"
#			],
#			....
#		},
#		"timers": {
#			"apt-daily": [
#				false,
#				"apt-daily.timer",
#				"loaded",
#				"active",
#				"waiting",
#				"Daily apt download activities"
#			],
#			....
#			"motd-news": [
#				false,
#				"motd-news.timer",
#				"loaded",
#				"active",
#				"waiting",
#				"Message of the Day"
#			],
#			....
#		}
#	}
#
def parse_systemctl_units(stdout:str, stderr:str, exitcode:int) -> dict:

	"""
	  UNIT                                                                                     LOAD      ACTIVE   SUB       DESCRIPTION
	  mono-xsp4.service                                                                        loaded    active   running   LSB: Mono XSP4
	  motd-news.service                                                                        loaded    inactive dead      Message of the Day
	● mountkernfs.service                                                                      masked    inactive dead      mountkernfs.service
	  systemd-machine-id-commit.service                                                        loaded    inactive dead      Commit a transient machine-id on disk
	● systemd-modules-load.service                                                             loaded    failed   failed    Load Kernel Modules
	  systemd-networkd-resolvconf-update.service                                               loaded    inactive dead      Update resolvconf for networkd DNS
	  sysinit.target                                                                           loaded    active   active    System Initialization
	● syslog.target                                                                            not-found inactive dead      syslog.target
	  time-sync.target                                                                         loaded    active   active    System Time Synchronized

	LOAD   = Reflects whether the unit definition was properly loaded.
	ACTIVE = The high-level unit activation state, i.e. generalization of SUB.
	SUB    = The low-level unit activation state, values depend on unit type.

	354 loaded units listed.
	To show all installed unit files use 'systemctl list-unit-files'.
	"""

	if exitcode != 0:
		raise Exception()

	# split into list of lines
	lines = LineList(stdout)
	assert isinstance(lines, LineList)

	# now we must separate a trailing description.
	lineNumbers = lines.getLineNumbersOfEmptyLines()
	assert lineNumbers
	assert lineNumbers[0] > 0
	del lines[lineNumbers[0]:]

	# get column split positions
	wordPos = [ 0 ] + getPositionsOfWords(lines[0])
	table = lines.createDataTableFromColumns(wordPos, bLStrip=True, bRStrip=True, bFirstLineIsHeader=True, columnDefs=[
		ColumnDef("MARK", _parseMark),
		ColumnDef("UNIT"),
		ColumnDef("LOAD"),
		ColumnDef("ACTIVE"),
		ColumnDef("SUB"),
		ColumnDef("DESCRIPTION"),
	])

	# build output matrix: use service names as keys
	ret = {}
	for record in table:
		key = record[1]
		pos = key.rfind(".")
		category = key[pos+1:] + "s"			# pluralize the category
		key = key[:pos]

		if category not in ret:
			ret[category] = {}

		ret[category][key] = record

	return ret
#



def get_systemctl_units(c = None) -> dict:
	stdout, stderr, exitcode = run(c, "/bin/systemctl -a list-units")
	return parse_systemctl_units(stdout, stderr, exitcode)
#























import collections


OSInfo = collections.namedtuple("OSInfo", [
	"class",
	"distribution",
	
])








#
# Returns:
#
#	{
#		"distribution": "ubuntu",
#		"lts": true,
#		"version": "16.04.6"
#	}
#
def get_etc_os_release(c = None) -> dict:
	stdout, stderr, exitcode = run(c, "cat /etc/os-release")
	return parse_etc_os_release(stdout, stderr, exitcode)
#














import re

from jk_cachefunccalls import cacheCalls

from .parsing_utils import *
from .invoke_utils import run



_parserKVPs = ParseAtFirstDelimiter(delimiter="=", valueCanBeWrappedInDoubleQuotes=True)



#
# Returns:
#
#	{
#		"distribution": "ubuntu",
#		"fullNameStr": "Ubuntu 16.04.6 LTS (Xenial Xerus)"
#		"lts": true,
#		"name": "Ubuntu"
#		"version": "16.04.6",
#		"versionID": "16.04",
#		"versionStr": "16.04.6 LTS (Xenial Xerus)",
#	}
#
def parse_etc_os_release(stdout:str, stderr:str, exitcode:int) -> dict:
	if exitcode != 0:
		raise Exception()

	data = _parserKVPs.parseLines(stdout.strip().split("\n"))
	ret = {
		"name": data["NAME"],									# "Ubuntu", "Raspbian GNU/Linux"
		"distribution": data["ID"],								# "ubuntu", "raspbian"
		"version": data["PRETTY_NAME"].split()[1],				# "16.04.6", "18.04.3", "10"
		"versionID": data["VERSION_ID"],						# "16.04", "18.04", "10"
		"versionStr": data["VERSION"],							# "16.04.6 LTS (Xenial Xerus)", "18.04.3 LTS (Bionic Beaver)", "10 (buster)"
		"lts": data["PRETTY_NAME"].endswith(" LTS"),			# True/False; might only work for Ubuntu
		"fullNameStr": data["NAME"] + " " + data["VERSION"],	# this should be the same as PRETTY_NAME
		"url": data["HOME_URL"],								# "https://www.ubuntu.com/"
	}

	m = re.match(r".+\s\(([^\)]+)\)$", ret["versionStr"])
	ret["codeName"] = m.group(1) if m else None

	return ret
#



#
# Returns:
#
#	{
#		"distribution": "ubuntu",
#		"fullNameStr": "Ubuntu 16.04.6 LTS (Xenial Xerus)"
#		"lts": true,
#		"name": "Ubuntu"
#		"version": "16.04.6",
#		"versionID": "16.04",
#		"versionStr": "16.04.6 LTS (Xenial Xerus)",
#	}
#
@cacheCalls(seconds=3, dependArgs=[0])
def get_etc_os_release(c = None) -> dict:
	stdout, stderr, exitcode = run(c, "cat /etc/os-release")
	return parse_etc_os_release(stdout, stderr, exitcode)
#















import os
import sys
import fabric

import jk_logging
import jk_sysinfo
import jk_json
import jk_argparsing
import jk_typing

from ..cli.ServerCfg import ServerCfg






class ElementarSysInfo(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# @param		str argLongOption			The arguments long option
	# @param		str argDescription			The arguments description
	# @param		callable function			The function to that will perform the data retrieval
	# @param		** extraArgs				Extra arguments to pass to the function as arguments
	#
	def __init__(self, argLongOption:str, argDescription:str, function, **kwargs):
		self.longOption = argLongOption
		self.description = argDescription
		self.function = function
		self.extraArgs = kwargs
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def onOptionEnable(self, argOption, argOptionArguments, parsedArgs):
		parsedArgs.optionData.set(self.longOption, True)
	#

	def onOptionDisable(self, argOption, argOptionArguments, parsedArgs):
		parsedArgs.optionData.set(self.longOption, False)
	#

	def register(self, ap:jk_argparsing.ArgsParser):
		ap.optionDataDefaults.set(self.longOption, False)
		c = ap.createOption(None, self.longOption, self.description + " (enable)").onOption = self.onOptionEnable
		c = ap.createOption(None, self.longOption + "-0", self.description + " (disable)").onOption = self.onOptionDisable
	#

	#
	# Perform data retrieval and return the result.
	#
	# @return		list|dict		The result data will be a JSON list or object.
	#
	def run(self, c):
		return self.function(c, **self.extraArgs)
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

#









ALL_SYSINFO_OPTIONS = [
	ElementarSysInfo("i-accesspoints", "Get data about access points", jk_sysinfo.get_accesspoints),
	ElementarSysInfo("i-apt-list-upgradable", "Get information about upgradable packages", jk_sysinfo.get_apt_list_upgradable),
	ElementarSysInfo("i-bios-info", "Get information about the BIOS", jk_sysinfo.get_bios_info),
	ElementarSysInfo("i-cpu-info", "Get information about the CPU", jk_sysinfo.get_cpu_info),
	ElementarSysInfo("i-date", "Get date (local)", jk_sysinfo.get_date, utc=False),
	ElementarSysInfo("i-date-utc", "Get date (UTC)", jk_sysinfo.get_date, utc=True),
	ElementarSysInfo("i-df", "Get disk space information", jk_sysinfo.get_df),
	ElementarSysInfo("i-docker-stats", "Retrieve information about running docker services", jk_sysinfo.get_docker_stats),
	ElementarSysInfo("i-dpkg-list", "Get information about installed packages", jk_sysinfo.get_dpkg_list),
	ElementarSysInfo("i-etc-hostname", "Get hostname information", jk_sysinfo.get_etc_hostname),
	ElementarSysInfo("i-etc-os-release", "Get information about the OS as stored in /etc/os/release", jk_sysinfo.get_etc_os_release),
	ElementarSysInfo("i-ifconfig", "Get information about the network interfaces as provided by 'ifconfig'", jk_sysinfo.get_ifconfig),
	ElementarSysInfo("i-lsb-release-a", "Get information about the current linux distribution", jk_sysinfo.get_lsb_release_a),
	ElementarSysInfo("i-lsblk", "Get information about block devices", jk_sysinfo.get_lsblk),
	ElementarSysInfo("i-lshw", "Get information about the hardware as provided by 'lshw'", jk_sysinfo.get_lshw),
	ElementarSysInfo("i-motherboard-info", "Get information about the motherboard", jk_sysinfo.get_motherboard_info),
	ElementarSysInfo("i-mount", "Get information about the mounted devices", jk_sysinfo.get_mount),
	ElementarSysInfo("i-needs-reboot", "Check if a reboot is required", jk_sysinfo.get_needs_reboot),
	ElementarSysInfo("i-net-info", "Retrieve various information about network devices", jk_sysinfo.get_net_info),
	ElementarSysInfo("i-pip3-list", "Retrieve a list of installed python packages", jk_sysinfo.get_pip3_list),
	ElementarSysInfo("i-proc-cpu-info", "Retrieve information about the CPU as provided by /proc/cpuinfo", jk_sysinfo.get_proc_cpu_info),
	ElementarSysInfo("i-proc-load-avg", "Retrieve information about the system load as provided by /proc/loadavg", jk_sysinfo.get_proc_load_avg),
	ElementarSysInfo("i-proc-meminfo", "Retrieve information about the memory usage as provided by /proc/meminfo", jk_sysinfo.get_proc_meminfo),
	ElementarSysInfo("i-ps", "Get information about processes", jk_sysinfo.get_ps),
	ElementarSysInfo("i-sensors", "Get information about processes", jk_sysinfo.get_sensors),
	ElementarSysInfo("i-systemctl-units", "Get information about all systemctrl units", jk_sysinfo.get_systemctl_units),
	ElementarSysInfo("i-uptime", "Get information about uptime", jk_sysinfo.get_uptime),
	ElementarSysInfo("i-user-info", "Get information about users", jk_sysinfo.get_user_info),
	ElementarSysInfo("i-vcgencmd", "Get information about a Raspberry Pi", jk_sysinfo.get_vcgencmd),
]
if os.getuid() == 0:
	ALL_SYSINFO_OPTIONS.extend([
		ElementarSysInfo("i-etc-group", "Get all data from /etc/group and /etc/gshadow", jk_sysinfo.get_etc_group),
		ElementarSysInfo("i-etc-passwd", "Get all data from /etc/passwd and /etc/shadow", jk_sysinfo.get_etc_passwd),
	])

ALL_SYSINFO_OPTIONS_MAP = {
	opt.longOption:opt for opt in ALL_SYSINFO_OPTIONS
}




class CLISysInfoJSON(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self):
		pass
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _onOptionAllAutoDetect(self, argOption, argOptionArguments, parsedArgs):
		for opt in ALL_SYSINFO_OPTIONS:
			if (opt.longOption == "i-vcgencmd") and not jk_sysinfo.has_local_vcgencmd():
				continue
			if (opt.longOption == "i-sensors") and not jk_sysinfo.has_local_sensors():
				continue
			if (opt.longOption == "i-docker-stats") and not jk_sysinfo.has_local_docker():
				continue
			parsedArgs.optionData.set(opt.longOption, True)
	#

	def _onOptionAllStd(self, argOption, argOptionArguments, parsedArgs):
		for opt in ALL_SYSINFO_OPTIONS:
			if opt.longOption in [ "i-vcgencmd" ]:
				continue
			parsedArgs.optionData.set(opt.longOption, True)
	#

	def _onOptionAllVM(self, argOption, argOptionArguments, parsedArgs):
		for opt in ALL_SYSINFO_OPTIONS:
			if opt.longOption in [ "i-vcgencmd" ]:
				continue
			if opt.longOption in [ "i-sensors" ]:
				continue
			parsedArgs.optionData.set(opt.longOption, True)
	#

	def _onOptionAllRPi(self, argOption, argOptionArguments, parsedArgs):
		for opt in ALL_SYSINFO_OPTIONS:
			if opt.longOption in []:
				continue
			parsedArgs.optionData.set(opt.longOption, True)
	#

	def _createArgsParser(self) -> jk_argparsing.ArgsParser:
		ap = jk_argparsing.ArgsParser(os.path.basename(__file__), "Display system information in JSON format.")

		ap.optionDataDefaults.set("help", False)
		ap.optionDataDefaults.set("color", True)
		ap.optionDataDefaults.set("outDirPath", None)
		ap.optionDataDefaults.set("promptPwdStdin", False)

		ap.createOption('h', 'help', "Display this help text.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("help", True)
		ap.createOption(None, 'color', "Force using colors in output.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("color", False)
		ap.createOption(None, 'no-color', "Don't use colors in output.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("color", False)
		ap.createOption(None, 'output-dir-path', "Write all data to individual JSON files in this directory instead of STDOUT.") \
			.expectDirectory("path", minLength=1, mustExist=True) \
			.onOption = \
				lambda argOption, argOptionArguments, parsedArgs: \
					parsedArgs.optionData.set("outDirPath", argOptionArguments[0])
		ap.createOption(None, 'remote', "The path of the configuration file for a remote host to connect to.") \
			.expectFile("file", toAbsolutePath=True, mustExist=True) \
			.onOption = lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("remote", argOptionArguments[0])
		ap.createOption(None, 'prompt-pwd-stdin', "If connecting to a remote host prompt for password via STDIN.").onOption = \
			lambda argOption, argOptionArguments, parsedArgs: \
				parsedArgs.optionData.set("promptPwdStdin", True)
		for opt in ALL_SYSINFO_OPTIONS:
			opt.register(ap)

		ap.createOption(None, 'i-all', "Use all system information modules (with autodetect).").onOption = self._onOptionAllAutoDetect
		ap.createOption(None, 'i-all-std', "Use all system information modules (Regular *nix Machines).").onOption = self._onOptionAllStd
		ap.createOption(None, 'i-all-rpi', "Use all system information modules (Raspberry Pi).").onOption = self._onOptionAllRPi
		ap.createOption(None, 'i-all-vm', "Use all system information modules (Virtual Machines).").onOption = self._onOptionAllVM

		ap.createAuthor("Jürgen Knauth", "jk@binary-overflow.de")
		ap.setLicense("Apache", YEAR = "2021-2023", COPYRIGHTHOLDER = "Jürgen Knauth")

		ap.createReturnCode(0, "Everything is okay.")
		ap.createReturnCode(1, "An error occurred.")

		return ap
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def main(self) -> int:
		ap = self._createArgsParser()

		# TODO: create logger with color autodetect here, as parsing might fail here and we should already have good log messages here
		parsedArgs = ap.parse()
		#parsedArgs.dump()

		if parsedArgs.optionData["help"]:
			ap.showHelp()
			return 0

		with jk_logging.wrapMain(bColor=parsedArgs.optionData["color"]) as log:
			#if not parsedArgs.optionData["color"]:
			#	log = jk_logging.ConsoleLogger.create(logMsgFormatter = jk_logging.DEFAULT_LOG_MESSAGE_FORMATTER, printToStdErr = True)

			"""
			cmdName, cmdArgs = parsedArgs.parseNextCommand()
			if cmdName is None:
				ap.showHelp()
				sys.exit(0)
			"""

			# ----

			c = None
			if "remote" in parsedArgs.optionData:
				_serverCfgFilePath = parsedArgs.optionData["remote"]
				serverCfg = ServerCfg.loadFromFile(_serverCfgFilePath)
				if serverCfg.shouldConnectViaUserNamePwd:
					if parsedArgs.optionData.get("promptPwdStdin"):
						serverCfg.readPwdFromSTDINIfPwdMissing()
					if serverCfg.pwd:
						log.notice("Connecting via username and password to " + serverCfg.host + " ...")
						c = fabric.Connection(host=serverCfg.host, user=serverCfg.user, port=serverCfg.port, connect_kwargs={"password": serverCfg.pwd})
					else:
						log.notice("Connecting via username to " + serverCfg.host + " ...")
						c = fabric.Connection(host=serverCfg.host, user=serverCfg.user, port=serverCfg.port)
				elif serverCfg.shouldConnectViaKeyFile:
					_keyFilePath = serverCfg.keyFile
					if not os.path.isabs(_keyFilePath):
						_keyFilePath = os.path.normpath(os.path.join(os.path.dirname(_serverCfgFilePath), _keyFilePath))
					log.notice("Connecting via keyfile to " + serverCfg.host + " ...")
					log.notice("Using key file: " + _keyFilePath)
					c = fabric.Connection(host=serverCfg.host, user=serverCfg.user, port=serverCfg.port, connect_kwargs={"key_filename": _keyFilePath})
				else:
					raise Exception()

			# ----

			bSuccess = True

			ret = {}
			for opt in ALL_SYSINFO_OPTIONS:
				if parsedArgs.optionData[opt.longOption]:
					try:
						ret[opt.longOption] = opt.run(c)
					except Exception as ee:
						# there has been an error
						ret[opt.longOption] = None
						# log.error("Failed to retrieve data for: " + opt.longOption)
						log.exception(ee)
						bSuccess = False

			# ----

			if not ret:
				ap.showHelp()
				return 0

			# ----

			outDirPath = parsedArgs.optionData["outDirPath"]
			if outDirPath:
				for k,v in ret.items():
					filePath = os.path.join(outDirPath, k + ".json")
					jk_json.saveToFilePretty(v, filePath)
			else:
				jk_json.prettyPrint(ret)

			# ----

			return 0 if bSuccess else 1
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

#











def main():
	app = CLISysInfoJSON()
	sys.exit(app.main())
#


















import os
import typing

import jk_typing
import jk_prettyprintobj

from .DrivePartitionInfo import DrivePartitionInfo






class DriveInfo(jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	@jk_typing.checkFunctionSignature()
	def __init__(self, jdata_lsblk_disk:dict, jdata_hdparam_I:dict):
		self.devicePath = jdata_lsblk_disk["dev"]
		self.diskGranularity = jdata_lsblk_disk["phy-sec"]
		self.isHotplug = jdata_lsblk_disk["hotplug"]
		self.model = jdata_lsblk_disk["model"]
		self.isReadOnly = jdata_lsblk_disk["ro"]
		self.isRotational = jdata_lsblk_disk["rota"]
		self.serial = jdata_lsblk_disk["serial"]
		self.size = jdata_lsblk_disk["size"]
		self.transport = jdata_lsblk_disk["tran"]
		self.uuid = jdata_lsblk_disk["uuid"]
		self.vendor = jdata_lsblk_disk["vendor"]

		self.formFactor = jdata_hdparam_I["configuration"]["formFactor"]
		self.nominalMediaRotationRate = jdata_hdparam_I["configuration"]["nominalMediaRotationRate"]
		self.transportHR = jdata_hdparam_I["general"]["transport"]
		self.firmwareRevision = jdata_hdparam_I["general"]["firmwareRevision"]

		#print(repr(jdata_hdparam_I["general"]["serial"]))
		#print(repr(self.serial))
		assert jdata_hdparam_I["general"]["serial"] == self.serial
		
		#print(repr(jdata_hdparam_I["general"]["model"]))
		#print(repr(self.model))
		#second model seems to be more reasonable - skip this test as there might be differences
		#assert jdata_hdparam_I["general"]["model"] == self.model

		self.isNCQSupported = "Native Command Queueing (NCQ)" in jdata_hdparam_I["features"]
		self.isTRIMSupported = False
		for k, v in jdata_hdparam_I["features"].items():
			if k.startswith("Data Set Management TRIM supported"):
				self.isTRIMSupported = v

		self.partitions = []
		if "children" in jdata_lsblk_disk:
			for jchild in jdata_lsblk_disk["children"]:
				self.partitions.append(DrivePartitionInfo(jchild))
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> list:
		return [
			"devicePath",
			"diskGranularity",
			"isHotplug",
			"model",
			"isReadOnly",
			"isRotational",
			"serial",
			"size",
			"transport",
			"uuid",
			"vendor",
			"formFactor",
			"nominalMediaRotationRate",
			"firmwareRevision",
			"transportHR",
			"isNCQSupported",
			"isTRIMSupported",
			"partitions",
		]
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def toJSON(self) -> dict:
		return {
			"devicePath": self.devicePath,
			"diskGranularity": self.diskGranularity,
			"isHotplug": self.isHotplug,
			"model": self.model,
			"isReadOnly": self.isReadOnly,
			"isRotational": self.isRotational,
			"serial": self.serial,
			"size": self.size,
			"transport": self.transport,
			"uuid": self.uuid,
			"vendor": self.vendor,
			"formFactor": self.formFactor,
			"nominalMediaRotationRate": self.nominalMediaRotationRate,
			"firmwareRevision": self.firmwareRevision,
			"transportHR": self.transportHR,
			"isNCQSupported": self.isNCQSupported,
			"isTRIMSupported": self.isTRIMSupported,
			"partitions": [
				x.toJSON() for x in self.partitions
			] if self.partitions else []
		}
	#

#












import os
import re
import pwd
import grp
import typing
import resource

# TODO: eliminate the use of modules pwd and grp as this way get_ps() can not be executed remotely.
#		NOTE: we might be able implement this using data from get_user_info().

from jk_cachefunccalls import cacheCalls
import jk_cmdoutputparsinghelper

from .parsing_utils import *
from .invoke_utils import run

_PAGESIZE = resource.getpagesize()
_PAGESIZE_KB = _PAGESIZE / 1024


_parserColonKVP = ParseAtFirstDelimiter(delimiter="=", valueCanBeWrappedInDoubleQuotes=True)



#
# Returns:
#
#	[
#			{
#					"args": "splash",
#					"cmd": "/sbin/init",
#					"pid": 1,
#					"ppid": 0,
#					"stat": "Ss",
#					"tty": null,
#					"uid": 0,
#					"user": "root"
#			},
#			{
#					"cmd": "[kthreadd]",
#					"pid": 2,
#					"ppid": 0,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			{
#					"cmd": "[ksoftirqd/0]",
#					"pid": 3,
#					"ppid": 2,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			...
#			{
#					"cmd": "bash",
#					"pid": 20144,
#					"ppid": 14839,
#					"stat": "Ss+",
#					"tty": "pts/3",
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			{
#					"args": "--spawner :1.9 /org/gtk/gvfs/exec_spaw/4",
#					"cmd": "/usr/lib/gvfs/gvfsd-computer",
#					"pid": 20292,
#					"ppid": 1,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#			{
#					"args": "/usr/share/code/resources/app/extensions/json-language-features/server/dist/jsonServerMain --node-ipc --clientProcessId=15491",
#					"cmd": "/usr/share/code/code",
#					"pid": 29554,
#					"ppid": 15491,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#	]
#
def parse_ps(stdout:str, stderr:str, exitcode:int, userIDToNameMap:typing.Dict[int,str] = None, grpIDToNameMap:typing.Dict[int,str] = None) -> dict:
	if exitcode != 0:
		raise Exception()

	lines = jk_cmdoutputparsinghelper.TextData(stdout).lines
	lines.removeTrailingEmptyLines()
	splitPositions = lines.identifySpaceColumnPositions(maxSplitPositions=6)
	table = lines.createDataTableFromColumns(positions=splitPositions, bLStrip=True, bRStrip=True, bFirstLineIsHeader=True, columnDefs=[
		jk_cmdoutputparsinghelper.ColumnDef("ppid", int),
		jk_cmdoutputparsinghelper.ColumnDef("pid", int),
		jk_cmdoutputparsinghelper.ColumnDef("tty", str),
		jk_cmdoutputparsinghelper.ColumnDef("stat", str),
		jk_cmdoutputparsinghelper.ColumnDef("uid", int),
		jk_cmdoutputparsinghelper.ColumnDef("gid", int),
		jk_cmdoutputparsinghelper.ColumnDef("cmd", str),
	])

	#	0	 1	 2		  3		 4		5	6
	# PPID   PID TT       STAT   UID   GID CMD

	ret = []
	for data in table.rowDictIterator():
		if data["tty"] == "?":
			data["tty"] = None

		try:
			# NOTE: this can only be read by root
			data["cwd"] = os.readlink("/proc/" + data["pid"] + "/cwd")
		except:
			pass

		_cmd = data["cmd"]
		pos = _cmd.find(" ")
		if pos > 0:
			data["cmd"] = _cmd[:pos]
			data["args"] = _cmd[pos+1:]

		if userIDToNameMap:
			data["user"] = userIDToNameMap.get(data["uid"], None)
		if grpIDToNameMap:
			data["group"] = grpIDToNameMap.get(data["gid"], None)

		ret.append(data)

	return ret
#



def _enrichWithVMSize(jProgramData:dict, pid:int):
	procPath = "/proc/" + str(jProgramData["pid"]) + "/statm"
	try:
		with open(procPath, "r") as f:
			components = f.read().strip().split(" ")
		assert len(components) == 7
		jProgramData["vmsizeKB"] = int(components[0]) * _PAGESIZE_KB
	except:
		jProgramData["vmsizeKB"] = -1
#




#
# Returns:
#
#	[
#			{
#					"args": "splash",
#					"cmd": "/sbin/init",
#					"pid": 1,
#					"ppid": 0,
#					"stat": "Ss",
#					"tty": null,
#					"uid": 0,
#					"user": "root"
#			},
#			{
#					"cmd": "[kthreadd]",
#					"pid": 2,
#					"ppid": 0,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			{
#					"cmd": "[ksoftirqd/0]",
#					"pid": 3,
#					"ppid": 2,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			...
#			{
#					"cmd": "bash",
#					"pid": 20144,
#					"ppid": 14839,
#					"stat": "Ss+",
#					"tty": "pts/3",
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			{
#					"args": "--spawner :1.9 /org/gtk/gvfs/exec_spaw/4",
#					"cmd": "/usr/lib/gvfs/gvfsd-computer",
#					"pid": 20292,
#					"ppid": 1,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#			{
#					"args": "/usr/share/code/resources/app/extensions/json-language-features/server/dist/jsonServerMain --node-ipc --clientProcessId=15491",
#					"cmd": "/usr/share/code/code",
#					"pid": 29554,
#					"ppid": 15491,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#	]
#
# NOTE: This method makes use of the python modules <c>pwd</c> and <c>grp</c> which make use of the local user and group database only.
#		For that reason this function can not be executed remotely until the implementation has been improved here.
#
def get_ps(c = None, bAddVMemSize:bool = False, userIDToNameMap:typing.Dict[int,str] = None, grpIDToNameMap:typing.Dict[int,str] = None) -> typing.List[dict]:
	stdout, stderr, exitcode = run(c, "ps ax -o ppid,pid,tty,stat,uid,gid,cmd")

	if c is None:
		if userIDToNameMap is None:
			userIDToNameMap = {}
			for entry in pwd.getpwall():
				userIDToNameMap[entry.pw_uid] = entry.pw_name
		if grpIDToNameMap is None:
			grpIDToNameMap = {}
			for entry in grp.getgrall():
				grpIDToNameMap[entry.gr_gid] = entry.gr_name

	ret = parse_ps(stdout, stderr, exitcode, userIDToNameMap, grpIDToNameMap)

	# remove ps process information itself and enrich this data
	ret2 = []
	for jProgramData in ret:
		if (jProgramData["cmd"] in [ "ps", "/bin/sh" ]) and (jProgramData["args"].find("ax -o ppid,pid,tty,stat,uid,gid,cmd") >= 0):
			continue
		if bAddVMemSize:
			_enrichWithVMSize(jProgramData, jProgramData["pid"])
		ret2.append(jProgramData)

	return ret2
#










import os
import re
import pwd
import grp
import typing
import resource

# TODO: eliminate the use of modules pwd and grp as this way get_ps() can not be executed remotely.
#		NOTE: we might be able implement this using data from get_user_info().

from jk_cachefunccalls import cacheCalls
import jk_cmdoutputparsinghelper

from .parsing_utils import *
from .invoke_utils import run

_PAGESIZE = resource.getpagesize()
_PAGESIZE_KB = _PAGESIZE / 1024


_parserColonKVP = ParseAtFirstDelimiter(delimiter="=", valueCanBeWrappedInDoubleQuotes=True)





def _enrichWithVMSize(jProgramData:dict, pid:int):
	procPath = "/proc/" + str(jProgramData["pid"]) + "/statm"
	try:
		with open(procPath, "r") as f:
			components = f.read().strip().split(" ")
		assert len(components) == 7
		jProgramData["vmsizeKB"] = int(components[0]) * _PAGESIZE_KB
	except:
		jProgramData["vmsizeKB"] = -1
#




def get_process_info_local(pid:int, userIDToNameMap:typing.Dict[int,str] = None, grpIDToNameMap:typing.Dict[int,str] = None):
	dirPath = "/proc/" + str(pid)

	#### /proc/????/status ####

	_fullContent = None
	try:
		with open(os.path.join(dirPath, "status"), "r") as f:
			_fullContent = f.read()
	except:
		return

	_pre_map = {}
	for line in _fullContent.split("\n"):
		if not line:
			continue
		m = re.match("^([a-zA-Z_]+):\s+(.+)$", line)
		if m is None:
			print("ERROR: Failed to parse line: " + line)
			return
		_pre_map[m.group(1)] = m.group(2)

	xxProcName = _pre_map["Name"]
	_sStateTemp = _pre_map["State"]
	m = re.match("^(.*)\s+\((.*)\)$", _sStateTemp)
	if m:
		xxState = m.group(1)
	else:
		print(_sStateTemp)
		xxState = "?"
	xxPPid = int(_pre_map["PPid"])
	_listUIDTemp = re.split("(\s+)", _pre_map["Uid"])
	_listGIDTemp = re.split("(\s+)", _pre_map["Gid"])
	_listUID = [ int(_listUIDTemp[0]), int(_listUIDTemp[2]), int(_listUIDTemp[4]), int(_listUIDTemp[6]) ]	# real, effective, saved, fs
	_listGID = [ int(_listGIDTemp[0]), int(_listGIDTemp[2]), int(_listGIDTemp[4]), int(_listGIDTemp[6]) ]	# real, effective, saved, fs
	xxVMSizeKB = None
	if "VmSize" in _pre_map:
		_vmSizeTemp = re.split("(\s+)", _pre_map["VmSize"])
		assert len(_vmSizeTemp) == 3
		assert _vmSizeTemp[2] == "kB"
		xxVMSizeKB = int(_vmSizeTemp[0])
	xxUID = _listUID[1]
	xxGID = _listGID[1]
	xxUserName = userIDToNameMap.get(xxUID, None) if userIDToNameMap else None
	xxGroupName = grpIDToNameMap.get(xxGID, None) if grpIDToNameMap else None

	#### /proc/????/cmdline ####

	xxCmd = None
	xxArgs = None
	with open(os.path.join(dirPath, "cmdline"), "r") as f:
		_cmdLine = f.read().rstrip("\x00").split("\x00")
		if _cmdLine:
			xxArgs = _cmdLine[1:]
			xxCmd = _cmdLine[0]
	if not xxCmd:
		assert xxProcName
		xxCmd = "[" + xxProcName + "]"

	#### /proc/????/cwd ####

	try:
		# NOTE: if we don't own this we can only read this as root
		xxCWD = os.readlink(os.path.join(dirPath, "cwd"))
	except:
		xxCWD = None

	#### /proc/????/exe ####

	try:
		# NOTE: if we don't own this we can only read this as root
		xxEXE = os.readlink(os.path.join(dirPath, "exe"))
	except:
		xxEXE = None

	# ----------------

	# sometimes xxArgs has not been provided with "\x00" but space as separator; compensate for this though this might not be perfect;
	if xxCmd and xxEXE:
		if xxCmd.startswith(xxEXE + " "):
			xxArgs = xxCmd[len(xxEXE)+1:].strip().split(" ")
			xxCmd = xxEXE

	assert xxCmd is not None
	assert xxArgs is not None

	# ----------------

	ret = {
		"args": " ".join(xxArgs),
		"argsv": xxArgs,
		"cmd": xxCmd,
		"pid": pid,
		"ppid": xxPPid,
		"stat": xxState,
	}
	if xxUserName is not None:
		ret["user"] = xxUserName
	if xxGroupName is not None:
		ret["group"] = xxGroupName
	if xxCWD is not None:
		ret["cwd"] = xxCWD
	if xxEXE is not None:
		ret["exe"] = xxEXE
	ret["vmsizeKB"] = xxVMSizeKB if xxVMSizeKB is not None else -1

	# ----------------

	return ret
#



#
# Returns:
#
#	[
#			{
#					"args": "splash",
#					"cmd": "/sbin/init",
#					"pid": 1,
#					"ppid": 0,
#					"stat": "Ss",
#					"tty": null,
#					"uid": 0,
#					"user": "root"
#			},
#			{
#					"cmd": "[kthreadd]",
#					"pid": 2,
#					"ppid": 0,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			{
#					"cmd": "[ksoftirqd/0]",
#					"pid": 3,
#					"ppid": 2,
#					"stat": "S",
#					"tty": null,
#					"uid": 0,
#					"user": "root",
#					"cwd": ....
#			},
#			...
#			{
#					"cmd": "bash",
#					"pid": 20144,
#					"ppid": 14839,
#					"stat": "Ss+",
#					"tty": "pts/3",
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			{
#					"args": "--spawner :1.9 /org/gtk/gvfs/exec_spaw/4",
#					"cmd": "/usr/lib/gvfs/gvfsd-computer",
#					"pid": 20292,
#					"ppid": 1,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#			{
#					"args": "/usr/share/code/resources/app/extensions/json-language-features/server/dist/jsonServerMain --node-ipc --clientProcessId=15491",
#					"cmd": "/usr/share/code/code",
#					"pid": 29554,
#					"ppid": 15491,
#					"stat": "Sl",
#					"tty": null,
#					"uid": 1000,
#					"user": "xxxxxxxx"
#					"gid": 1000,
#					"group": "xxxxxxxx",
#					"cwd": ....
#			},
#			...
#	]
#
def get_ps_local(c = None, bAddVMemSize:bool = False, userIDToNameMap:typing.Dict[int,str] = None, grpIDToNameMap:typing.Dict[int,str] = None) -> typing.List[dict]:
	if c is not None:
		raise Exception("This process analysis requires fast I/O and therefore can only be run locally!")

	if c is None:
		if userIDToNameMap is None:
			userIDToNameMap = {}
			for entry in pwd.getpwall():
				userIDToNameMap[entry.pw_uid] = entry.pw_name
		if grpIDToNameMap is None:
			grpIDToNameMap = {}
			for entry in grp.getgrall():
				grpIDToNameMap[entry.gr_gid] = entry.gr_name

	# ----

	ret = []
	for fe in os.scandir("/proc"):
		if fe.is_dir(follow_symlinks=False) and re.match("^[1-9][0-9]*$", fe.name):
			ret.append(get_process_info_local(int(fe.name), userIDToNameMap, grpIDToNameMap))

	# ----

	return ret
#





